import unittest
import os
import os.path
import pytest
import pandas as pd

import sklearn.metrics
from sklearn.datasets import load_boston

from analitico.factory import Factory
from analitico.plugin import *
from .test_mixin import TestMixin

# pylint: disable=no-member

ASSETS_PATH = os.path.dirname(os.path.realpath(__file__)) + "/assets"


@pytest.mark.django_db
class CatBoostTests(unittest.TestCase, TestMixin):
    """ Unit testing of machine learning algorithms """

    def train_iris(self, factory):
        """ Train iris.csv dataset using a multiclass classifier, return df and training results """
        csv_path = self.get_asset_path("iris_1.csv")
        df = pd.read_csv(csv_path)
        df = df.drop(columns=["Id"])
        df["Species"] = df["Species"].astype("category")
        catboost = CatBoostPlugin(factory=factory, parameters={"learning_rate": 0.2})
        # run training
        training = catboost.run(df.copy(), action="recipe/train")
        self.assertIsNotNone(training)
        return df, training

    def test_catboost_binary_classifier_training(self):
        """ Test training catboost as a binary classifier """
        try:
            with Factory() as factory:
                csv_path = self.get_asset_path("titanic_1.csv")
                recipe = RecipePipelinePlugin(
                    factory=factory,
                    plugins=[
                        CsvDataframeSourcePlugin(
                            source={"url": csv_path, "schema": {"apply": [{"name": "Survived", "type": "category"}]}}
                        ),
                        CatBoostPlugin(parameters={"learning_rate": 0.2}),
                    ],
                )

                self.assertEqual(len(recipe.plugins), 2)
                self.assertTrue(isinstance(recipe, RecipePipelinePlugin))
                self.assertTrue(isinstance(recipe.plugins[0], CsvDataframeSourcePlugin))
                self.assertTrue(isinstance(recipe.plugins[1], CatBoostPlugin))

                # run training
                results = recipe.run(action="recipe/train")

                self.assertIsNotNone(results)
                self.assertEqual(results["data"]["label"], "Survived")
                self.assertEqual(results["data"]["source_records"], 891)
                self.assertEqual(results["data"]["training_records"], 712)
                self.assertEqual(results["data"]["test_records"], 179)

                self.assertEqual(len(results["data"]["classes"]), 2)
                self.assertEqual(results["data"]["classes"][0], 0)
                self.assertEqual(results["data"]["classes"][1], 1)

                self.assertEqual(results["parameters"]["loss_function"], "Logloss")
                self.assertEqual(results["parameters"]["test_size"], 0.2)
                self.assertEqual(results["parameters"]["learning_rate"], 0.2)

                # model was saved?
                artifacts = factory.get_artifacts_directory()
                model_path = os.path.join(artifacts, "model.cbm")
                self.assertTrue(os.path.isfile(model_path))

        except Exception as exc:
            factory.error("test_catboost_binary_classifier - " + str(exc))
            pass

    def test_catboost_binary_classifier_prediction(self):
        """ Test predictions with catboost as a binary classifier """
        try:
            with Factory() as factory:
                csv_path = self.get_asset_path("titanic_1.csv")
                df = pd.read_csv(csv_path)
                df["Survived"] = df["Survived"].astype("category")
                catboost = CatBoostPlugin(factory=factory, parameters={"learning_rate": 0.2})

                # run training
                training = catboost.run(df.copy(), action="recipe/train")
                self.assertIsNotNone(training)

                df_labels = df[["Survived"]]
                df = df.drop(columns=["Survived"])
                predict = catboost.run(df, action="endpoint/predict")

                # check to make sure predictions are from available labels
                self.assertIn("predictions", predict)
                self.assertEqual(len(predict["predictions"]), len(df))
                for prediction in predict["predictions"]:
                    self.assertIn(prediction, training["data"]["classes"])

                # make sure each record had each class scored
                self.assertIn("probabilities", predict)
                self.assertEqual(len(predict["probabilities"]), len(df))
                for probability in predict["probabilities"]:
                    for label_class in training["data"]["classes"]:
                        self.assertIn(label_class, probability)

                # check correctness of predictions
                report = sklearn.metrics.classification_report(df_labels, predict["predictions"], output_dict=True)
                self.assertGreater(report["0"]["f1-score"], 0.80)
                self.assertGreater(report["1"]["f1-score"], 0.65)

        except Exception as exc:
            factory.error("test_catboost_multiclass_classifier_prediction - " + str(exc))
            raise exc

    def test_catboost_binary_classifier_prediction_with_labels(self):
        """ Test predictions with catboost as a binary classifier (using labels instead of int) """
        try:
            with Factory() as factory:
                csv_path = self.get_asset_path("titanic_1.csv")
                df = pd.read_csv(csv_path)
                df["Survived"] = df["Survived"].map({0: "No", 1: "Yes"})
                df["Survived"] = df["Survived"].astype("category")

                catboost = CatBoostPlugin(factory=factory, parameters={"learning_rate": 0.2})

                # run training
                training = catboost.run(df.copy(), action="recipe/train")
                self.assertIsNotNone(training)

                df_labels = df[["Survived"]]
                df = df.drop(columns=["Survived"])
                predict = catboost.run(df, action="endpoint/predict")

                # check to make sure predictions are from available labels
                self.assertIn("predictions", predict)
                self.assertEqual(len(predict["predictions"]), len(df))
                for prediction in predict["predictions"]:
                    self.assertIn(prediction, training["data"]["classes"])

                # make sure each record had each class scored
                self.assertIn("probabilities", predict)
                self.assertEqual(len(predict["probabilities"]), len(df))
                for probability in predict["probabilities"]:
                    for label_class in training["data"]["classes"]:
                        self.assertIn(label_class, probability)

                # check correctness of predictions
                report = sklearn.metrics.classification_report(df_labels, predict["predictions"], output_dict=True)
                self.assertGreater(report["No"]["f1-score"], 0.80)
                self.assertGreater(report["Yes"]["f1-score"], 0.65)

        except Exception as exc:
            factory.error("test_catboost_multiclass_classifier_prediction - " + str(exc))
            pass

    def test_catboost_multiclass_classifier_training(self):
        """ Test training catboost as a multiclass classifier """
        try:
            with Factory() as factory:
                df, results = self.train_iris(factory)

                self.assertEqual(results["data"]["label"], "Species")
                self.assertEqual(results["data"]["source_records"], 150)
                self.assertEqual(results["data"]["training_records"], 120)
                self.assertEqual(results["data"]["test_records"], 30)

                self.assertEqual(len(results["data"]["classes"]), 3)
                self.assertEqual(results["data"]["classes"][0], "Iris-setosa")
                self.assertEqual(results["data"]["classes"][1], "Iris-versicolor")
                self.assertEqual(results["data"]["classes"][2], "Iris-virginica")

                self.assertEqual(results["parameters"]["loss_function"], "MultiClass")
                self.assertEqual(results["parameters"]["test_size"], 0.2)
                self.assertEqual(results["parameters"]["learning_rate"], 0.2)
                self.assertEqual(results["parameters"]["iterations"], 50)

                self.assertEqual(results["scores"]["accuracy_score"], 1.0)
                self.assertLess(results["scores"]["log_loss"], 0.10)

                self.assertEqual(len(results["scores"]["features_importance"]), 4)
                self.assertIn("PetalLengthCm", results["scores"]["features_importance"])
                self.assertIn("PetalWidthCm", results["scores"]["features_importance"])
                self.assertIn("SepalLengthCm", results["scores"]["features_importance"])
                self.assertIn("SepalWidthCm", results["scores"]["features_importance"])

                # model was saved?
                artifacts = factory.get_artifacts_directory()
                model_path = os.path.join(artifacts, "model.cbm")
                self.assertTrue(os.path.isfile(model_path))

        except Exception as exc:
            factory.error("test_catboost_multiclass_classifier - " + str(exc))
            pass

    def test_catboost_multiclass_classifier_training_classification_report(self):
        try:
            with Factory() as factory:
                df, results = self.train_iris(factory)

                classes = results["data"]["classes"]
                scores = results["scores"]
                self.assertIn("classification_report", scores)
                for class_name in classes:
                    self.assertIn(class_name, scores["classification_report"])
                    class_report = scores["classification_report"][class_name]
                    self.assertIn("f1-score", class_report)
                    self.assertIn("precision", class_report)
                    self.assertIn("recall", class_report)
                    self.assertIn("support", class_report)

                self.assertIn("confusion_matrix", scores)
                confusion_matrix = scores["confusion_matrix"]
                self.assertEqual(len(confusion_matrix), len(classes))
                for line in confusion_matrix:
                    self.assertEqual(len(line), len(classes))

        except Exception as exc:
            factory.error("test_catboost_multiclass_classifier_training_classification_report - " + str(exc))
            pass

    def test_catboost_multiclass_classifier_prediction(self):
        """ Test predictions with catboost as a binary classifier """
        try:
            with Factory() as factory:
                df, training = self.train_iris(factory)

                df_labels = df[["Species"]]
                df = df.drop(columns=["Species"])
                predict = catboost.run(df, action="endpoint/predict")

                # check to make sure predictions are from available labels
                self.assertIn("predictions", predict)
                self.assertEqual(len(predict["predictions"]), len(df))
                for prediction in predict["predictions"]:
                    self.assertIn(prediction, training["data"]["classes"])

                # make sure each record had each class scored
                self.assertIn("probabilities", predict)
                self.assertEqual(len(predict["probabilities"]), len(df))
                for probability in predict["probabilities"]:
                    for label_class in training["data"]["classes"]:
                        self.assertIn(label_class, probability)

                # check correctness of predictions
                # https://scikit-learn.org/stable/modules/generated/sklearn.metrics.classification_report.html
                report = sklearn.metrics.classification_report(df_labels, predict["predictions"], output_dict=True)
                self.assertGreater(report["Iris-setosa"]["f1-score"], 0.95)
                self.assertGreater(report["Iris-versicolor"]["f1-score"], 0.95)
                self.assertGreater(report["Iris-virginica"]["f1-score"], 0.95)

        except Exception as exc:
            factory.error("test_catboost_multiclass_classifier_prediction - " + str(exc))
            pass

    def test_catboost_regressor_training(self):
        """ Test training catboost as a regressor """
        try:
            with Factory() as factory:
                # boston data info here:
                # https://towardsdatascience.com/linear-regression-on-boston-housing-dataset-f409b7e4a155
                # bare bones, just run the plugin by itself w/o pipeline
                catboost = CatBoostPlugin(factory=factory, parameters={"learning_rate": 0.2})

                boston_dataset = load_boston()
                boston = pd.DataFrame(boston_dataset.data, columns=boston_dataset.feature_names)
                boston["MEDV"] = boston_dataset.target

                results = catboost.run(boston, action="recipe/train")

                self.assertIsNotNone(results)
                self.assertEqual(results["data"]["label"], "MEDV")  # median value
                self.assertEqual(results["data"]["source_records"], 506)
                self.assertEqual(results["data"]["training_records"], 404)
                self.assertEqual(results["data"]["test_records"], 102)

                # not a classifier
                self.assertFalse("classes" in results["data"])
                self.assertFalse("accuracy_score" in results["scores"])
                self.assertFalse("log_loss" in results["scores"])

                self.assertEqual(results["parameters"]["loss_function"], "RMSE")
                self.assertEqual(results["parameters"]["test_size"], 0.2)
                self.assertEqual(results["parameters"]["learning_rate"], 0.2)
                self.assertEqual(results["parameters"]["iterations"], 50)

                self.assertEqual(len(results["scores"]["features_importance"]), 13)
                self.assertGreater(results["scores"]["features_importance"]["AGE"], 8.0)
                self.assertGreater(results["scores"]["features_importance"]["DIS"], 8.0)
                self.assertGreater(results["scores"]["features_importance"]["LSTAT"], 20.0)

                self.assertIn("mean_abs_error", results["scores"])
                self.assertIn("median_abs_error", results["scores"])
                self.assertIn("sqrt_mean_squared_error", results["scores"])
                self.assertIn("mean_abs_error", results["scores"])

                # model was saved?
                artifacts = factory.get_artifacts_directory()
                model_path = os.path.join(artifacts, "model.cbm")
                self.assertTrue(os.path.isfile(model_path))

        except Exception as exc:
            factory.error("test_catboost_regressor - " + str(exc))
            pass

    def test_catboost_regressor_prediction(self):
        """ Test predictions with catboost as a regressor """
        try:
            with Factory() as factory:
                boston_dataset = load_boston()
                boston = pd.DataFrame(boston_dataset.data, columns=boston_dataset.feature_names)
                boston["MEDV"] = boston_dataset.target

                catboost = CatBoostPlugin(factory=factory, parameters={"learning_rate": 0.2})
                training = catboost.run(boston, action="recipe/train")
                self.assertIsNotNone(training)

                boston_dataset = load_boston()
                boston = pd.DataFrame(boston_dataset.data, columns=boston_dataset.feature_names)
                boston_labels = boston_dataset.target
                predict = catboost.run(boston, action="endpoint/predict")

                # check to make sure predictions are from available labels
                self.assertIn("predictions", predict)
                self.assertEqual(len(predict["predictions"]), len(boston))

                # https://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_absolute_error.html
                mae = sklearn.metrics.mean_absolute_error(boston_labels, predict["predictions"])
                self.assertLessEqual(mae, 3)

        except Exception as exc:
            factory.error("test_catboost_regressor_prediction - " + str(exc))
            pass



import unittest

import analitico.mixin


class MyClass1(analitico.mixin.AttributeMixin):
    """ Class has no __init__ method """

    pass


class MyClass2(analitico.mixin.AttributeMixin):
    """ Class with passthrough init method """

    mickey = "empty"

    def __init__(self, mickey, *args, **kwargs):
        # here we call super().__init__ correctly and pass down the attributes
        super().__init__(*args, **kwargs)
        self.mickey = mickey


class MyClass3(analitico.mixin.AttributeMixin):
    """ Class with no passthrough init method """

    mickey = "empty"

    def __init__(self, mickey, **kwargs):
        # here we don't call super().__init__ and loose all attributes but mickey's
        self.mickey = mickey


class AttributesMixinTests(unittest.TestCase):
    """ Unit testing for mixin classes """

    def test_mixin_attributes(self):
        """ Basic basic functionality of AttributeMixin """
        obj1 = MyClass1(mickey="smart", goofy="funny", minnie="cute")

        self.assertEqual(obj1.mickey, "smart")
        self.assertEqual(obj1.goofy, "funny")
        self.assertEqual(obj1.minnie, "cute")

    def test_mixin_attributes_missing_attribute(self):
        """ Basic missing attribute in AttributeMixin """
        obj1 = MyClass1(mickey="smart", goofy="funny", minnie="cute")

        with self.assertRaises(AttributeError):
            obj1.pinocchio

    def test_mixin_attributes_init_passthrough(self):
        """ Basic passthrough initialization of AttributeMixin """
        obj1 = MyClass2(mickey="smart", goofy="funny", minnie="cute")

        self.assertEqual(obj1.mickey, "smart")
        self.assertEqual(obj1.goofy, "funny")
        self.assertEqual(obj1.minnie, "cute")

        # bogus attribute not defined
        with self.assertRaises(AttributeError):
            obj1.pinocchio

    def test_mixin_attributes_init_no_passthrough(self):
        """ Basic passthrough initialization of AttributeMixin """
        obj1 = MyClass3(mickey="smart", goofy="funny", minnie="cute")
        self.assertEqual(obj1.mickey, "smart")

        # no super().__init__ so these are not defined
        with self.assertRaises(AttributeError):
            self.assertEqual(obj1.goofy, "funny")
        with self.assertRaises(AttributeError):
            self.assertEqual(obj1.minnie, "cute")

        # bogus attribute also not defined
        with self.assertRaises(AttributeError):
            obj1.pinocchio



import analitico.utilities
import pandas as pd
import os

from analitico.schema import generate_schema
from .pipelineplugin import PipelinePlugin
from .interfaces import plugin

##
## DataframePipelinePlugin
##


@plugin
class DataframePipelinePlugin(PipelinePlugin):
    """ 
    A ETL pipeline plugin that creates a linear workflow by chaining together other plugins 
    where the final result is a pandas dataframe + its schema (metadata). These get saved
    as artifacts named data.csv (the data) and data.csv.info (the schema).
    """

    class Meta(PipelinePlugin.Meta):
        name = "analitico.plugin.DataframePipelinePlugin"
        inputs = None
        outputs = [{"name": "dataframe", "type": "pandas.DataFrame"}]

    def run(self, *args, action=None, **kwargs):
        """ Process the plugins in sequence then save the resulting dataframe """
        df = super().run(*args, action=action, **kwargs)
        if not isinstance(df, pd.DataFrame):
            self.logger.warn("DataframePipelinePlugin.run - pipeline didn't produce a valid dataframe")
            return None

        # save dataframe as data.csv
        # we will save the index column only if it is named
        # and it was created explicitely
        artifacts_path = self.factory.get_artifacts_directory()
        csv_path = os.path.join(artifacts_path, "data.csv")
        index = bool(df.index.name)
        df.to_csv(csv_path, index=index)

        # save schema as data.csv.info
        schema = generate_schema(df)
        csv_info_path = csv_path + ".info"
        analitico.utilities.save_json({"schema": schema}, csv_info_path)

        return df



import os
import pandas as pd
import tempfile
import urllib
import requests
import base64

from analitico import AnaliticoException
from analitico.mixin import AttributeMixin
from analitico.utilities import save_text, subprocess_run

from collections import OrderedDict
from pathlib import Path


class Item(AttributeMixin):
    """ Base class for items like datasets, recipes and notebooks on Analitico. """

    # SDK used to communicate with the service
    sdk = None

    def __init__(self, sdk, item_data: dict):
        # item's basics
        self.sdk = sdk
        self.id = item_data["id"]
        self.type = item_data["type"]
        # big bag of open ended properties to be retrieved via get_attribute
        super().__init__(**item_data["attributes"])

    ##
    ## Properties
    ##

    @property
    def title(self) -> str:
        """ The title of this item. """
        return self.get_attribute("title")

    @title.setter
    def title(self, title: str):
        """ Set the title of this item. """
        self.set_attribute("title", title)

    @property
    def url(self):
        """ Url of this item on the service. """
        item_url = f"analitico://{self.sdk.get_item_type(self.id)}s/{self.id}"
        item_url, _ = self.sdk.get_url_headers(item_url)
        return item_url

    ##
    ## Methods
    ##

    def upload(self, filepath: str = None, df: pd.DataFrame = None, remotepath: str = None) -> bool:

        if isinstance(df, pd.DataFrame):
            if not filepath:
                filepath = "data.parquet"

            if not isinstance(filepath, Path):
                filepath = Path(filepath)

            supported_formats = (".parquet", ".csv")
            format = filepath.suffix
            if format not in supported_formats:
                raise AnaliticoException(
                    f"Supported formats for automatic dataframe encoding are: {str(supported_formats)}"
                )

            # encode dataframe to disk temporarily
            with tempfile.NamedTemporaryFile(mode="w+", prefix="df_", suffix=format) as f:
                if format == ".parquet":
                    df.to_parquet(f.name)
                elif format == ".csv":
                    df.to_csv(f.name)
                # upload single file, don't use the name of temp file
                return self.upload(filepath=f.name, remotepath=filepath)

        # need to specify a file, directory or Path that should be uploaded to this item's storage
        if not isinstance(filepath, str) and not isinstance(filepath, Path):
            raise AnaliticoException("Please provide a path to the file or files to be uploaded.")

        if not remotepath:
            remotepath = filepath

        # uploading a single file?
        if os.path.isfile(filepath):
            # retrieve workspace associated with this item
            workspace_id = self.get_attribute("workspace_id")
            workspace = self.sdk.get_workspace(workspace_id)

            # storage configuration for this workspace
            storage = workspace.get_attribute("storage")
            assert storage["driver"] == "hetzner-webdav"
            username = storage["credentials"]["username"]
            password = storage["credentials"]["password"]

            if True:
                # TODO move driver to public repo and use mkdirs
                # TODO supports only files in base directory, should support recursive
                remote_url = f"{storage['url']}/{self.sdk.get_item_type(self.id)}s/"
                response = requests.request("MKCOL", remote_url, auth=(username, password))
                remote_url += f"{self.id}/"
                response = requests.request("MKCOL", remote_url, auth=(username, password))

                with open(filepath, "rb") as f:
                    remote_url += str(remotepath)
                    response = requests.put(remote_url, data=f, auth=(username, password))
                    assert response.status_code == 200 or response.status_code == 201
                    return True

            if False:
                rsync_path = f"{username}@{username}.your-storagebox.de"
#                rsync_path = f"{rsync_path}:./{self.sdk.get_item_type(self.id)}s/{self.id}/{remotepath}"
                rsync_path = f"{rsync_path}:./pippo_{remotepath}"

                with tempfile.NamedTemporaryFile(suffix=".id_rsa") as key:
                    signature = storage["credentials"]["ssh_private_key"]
                    signature = str(base64.b64decode(signature), "ascii")
                    save_text(signature, key.name)
                    os.chmod(key.name, 0o600)

                    with tempfile.NamedTemporaryDirectory() as tempdir:
                        # create relative paths
                        pass

                    sync_cmd = [
                        "rsync",
                        "--recursive",
                        "--progress",
                        "-L", # follow symlinks
                        "-e",
                        f"'ssh -p23 -o StrictHostKeyChecking=no -i {key.name}'",
                        filepath,
                        rsync_path,
                    ]
                    cmd = " ".join(sync_cmd)
                    os.system(cmd)

                    # a,b = subprocess_run(sync_cmd)

        raise NotImplementedError("Uploading multiple files at once is not yet implemented.")

    def save(self) -> bool:
        """ Save any changes to the service. """
        json = self.sdk.get_url_json(self.url, method="PUT", json=self.to_dict(), status_code=200)
        self.attributes = json["data"]["attributes"]
        return True

    def delete(self) -> bool:
        """
        Delete this item from the service.
        
        Returns:
            bool -- True if item was deleted.
        """
        self.sdk.get_url_json(self.url, method="DELETE", status_code=204)
        return True

    def to_dict(self) -> dict:
        """ Return item as a dictionary. """
        return {"id": self.id, "type": self.type, "attributes": self.attributes}



# Regression and classification plugins based on CatBoost

import pandas as pd
import numpy as np
import os.path

import sklearn.metrics
from sklearn.model_selection import train_test_split
from sklearn.metrics import (
    mean_squared_error,
    mean_absolute_error,
    median_absolute_error,
    accuracy_score,
    precision_score,
    recall_score,
    classification_report,
    confusion_matrix,
)

import catboost
from catboost import CatBoostClassifier, CatBoostRegressor

from analitico.utilities import time_ms

import analitico.pandas
import analitico.schema
from analitico.schema import generate_schema, ANALITICO_TYPE_CATEGORY, ANALITICO_TYPE_INTEGER, ANALITICO_TYPE_FLOAT
from .interfaces import (
    IAlgorithmPlugin,
    PluginError,
    plugin,
    ALGORITHM_TYPE_REGRESSION,
    ALGORITHM_TYPE_BINARY_CLASSICATION,
    ALGORITHM_TYPE_MULTICLASS_CLASSIFICATION,
)

##
## CatBoostPlugin
##


@plugin
class CatBoostPlugin(IAlgorithmPlugin):
    """ Base class for CatBoost regressor and classifier plugins """

    results = None

    class Meta(IAlgorithmPlugin.Meta):
        name = "analitico.plugin.CatBoostPlugin"
        algorithms = [
            ALGORITHM_TYPE_REGRESSION,
            ALGORITHM_TYPE_BINARY_CLASSICATION,
            ALGORITHM_TYPE_MULTICLASS_CLASSIFICATION,
        ]

    def create_model(self, results):
        """ Creates actual CatBoostClassifier or CatBoostRegressor model """
        iterations = self.get_attribute("parameters.iterations", 50)
        learning_rate = self.get_attribute("parameters.learning_rate", 1)
        depth = self.get_attribute("parameters.depth", 8)
        if results:
            results["parameters"]["iterations"] = iterations
            results["parameters"]["learning_rate"] = learning_rate
            results["parameters"]["depth"] = depth

        algo = results.get("algorithm", ALGORITHM_TYPE_REGRESSION)
        if algo == ALGORITHM_TYPE_REGRESSION:
            return CatBoostRegressor(iterations=iterations, learning_rate=learning_rate, depth=depth)
        elif algo == ALGORITHM_TYPE_BINARY_CLASSICATION:
            # task_type="GPU", # runtime will pick up the GPU even if we don't specify it here
            return CatBoostClassifier(
                iterations=iterations, learning_rate=learning_rate, depth=depth, loss_function="Logloss"
            )
        elif algo == ALGORITHM_TYPE_MULTICLASS_CLASSIFICATION:
            return CatBoostClassifier(
                iterations=iterations, learning_rate=learning_rate, depth=depth, loss_function="MultiClass"
            )
        else:
            raise PluginError("CatBoostPlugin.create_model - can't handle algorithm type: %s", results["algorithm"])

    def get_categorical_idx(self, df):
        """ Return indexes of the columns that should be considered categorical for the purpose of catboost training """
        categorical_idx = []
        for i, column in enumerate(df.columns):
            if analitico.schema.get_column_type(df, column) is analitico.schema.ANALITICO_TYPE_CATEGORY:
                categorical_idx.append(i)
                df[column].replace(np.nan, "", regex=True, inplace=True)
                self.factory.debug("%3d %s (%s/categorical)", i, column, df[column].dtype.name)
            else:
                self.factory.debug("%3d %s (%s)", i, column, df[column].dtype.name)
        return categorical_idx

    def validate_schema(self, train_df, test_df):
        """ Checks training and test dataframes to make sure they have matching schemas """
        train_schema = generate_schema(train_df)
        if test_df:
            test_schema = generate_schema(test_df)
            train_columns = train_schema["columns"]
            test_columns = test_schema["columns"]
            if len(train_columns) != len(test_columns):
                msg = "{} - training data has {} columns while test data has {} columns".format(
                    self.name, len(train_columns), len(test_columns)
                )
                raise PluginError(msg)
            for i in range(0, len(train_columns)):
                if train_columns[i]["name"] != test_columns[i]["name"]:
                    msg = "{} - column {} of train '{}' and test '{}' have different names".format(
                        self.name, i, train_columns[i]["name"], test_columns[i]["name"]
                    )
                    raise PluginError(msg)
                if train_columns[i]["type"] != test_columns[i]["type"]:
                    msg = "{} - column {} of train '{}' and test '{}' have different names".format(
                        self.name, i, train_columns[i]["type"], test_columns[i]["type"]
                    )
                    raise PluginError(msg)
        return train_schema

    def score_training(
        self,
        model: catboost.CatBoost,
        test_df: pd.DataFrame,
        test_pool: catboost.Pool,
        test_labels: pd.DataFrame,
        results: dict,
    ):
        """ Scores the results of this training """
        for key, value in model.get_params().items():
            results["parameters"][key] = value

        results["scores"]["best_iteration"] = model.get_best_iteration()

        best_score = model.get_best_score()
        try:
            best_score["training"] = best_score.pop("learn")
            best_score["validation"] = best_score.pop("validation_0")
        except KeyError:
            pass
        results["scores"]["best_score"] = best_score

        # result for each evaluation epoch
        evals_result = model.get_evals_result()
        try:
            evals_result["training"] = evals_result.pop("learn")
            evals_result["validation"] = evals_result.pop("validation_0")
        except KeyError:
            pass
        results["scores"]["iterations"] = evals_result

        # catboost can tell which features weigh more heavily on the predictions
        self.info("features importance:")
        features_importance = results["scores"]["features_importance"] = {}
        for label, importance in model.get_feature_importance(prettified=True):
            features_importance[label] = round(importance, 5)
            self.info("%24s: %8.4f", label, importance)

        # make the prediction using the resulting model
        # output test set with predictions
        # after moving label to the end for easier reading
        test_predictions = model.predict(test_pool)
        label = test_labels.name
        test_df[label] = test_labels
        cols = list(test_df.columns.values)
        cols.pop(cols.index(label))
        test_df = test_df[cols + [label]]
        test_df["prediction"] = test_predictions
        test_df = analitico.pandas.pd_sample(test_df, 200)  # just sampling
        artifacts_path = self.factory.get_artifacts_directory()
        test_df.to_csv(os.path.join(artifacts_path, "test.csv"))

    def score_regressor_training(self, model, test_df, test_pool, test_labels, results):
        test_preds = model.predict(test_pool)
        results["scores"]["median_abs_error"] = round(median_absolute_error(test_preds, test_labels), 5)
        results["scores"]["mean_abs_error"] = round(mean_absolute_error(test_preds, test_labels), 5)
        results["scores"]["sqrt_mean_squared_error"] = round(np.sqrt(mean_squared_error(test_preds, test_labels)), 5)

    def score_classifier_training(self, model, test_df, test_pool, test_labels, results):
        """ Scores the results of this training for the CatBoostClassifier model """
        # There are many metrics available:
        # https://scikit-learn.org/stable/modules/classes.html#module-sklearn.metrics

        scores = results["scores"]
        train_classes = results["data"]["classes"]  # the classes (actual strings)
        train_classes_codes = list(range(0, len(train_classes)))  # the codes, eg: 0, 1, 2...

        test_true = list(test_labels)  # test true labels
        test_preds = model.predict(test_pool, prediction_type="Class")  # prediction for each test sample
        test_probs = model.predict_proba(test_pool, verbose=True)  # probability for each class for each sample

        # Log loss, aka logistic loss or cross-entropy loss.
        scores["log_loss"] = round(sklearn.metrics.log_loss(test_true, test_probs, labels=train_classes_codes), 5)

        # In multilabel classification, this function computes subset accuracy:
        # the set of labels predicted for a sample must exactly match the corresponding set of labels in y_true.
        scores["accuracy_score"] = round(accuracy_score(test_true, test_preds), 5)

        # The precision is the ratio tp / (tp + fp) where tp is the number of true positives
        # and fp the number of false positives. The precision is intuitively the ability
        # of the classifier not to label as positive a sample that is negative.
        # The best value is 1 and the worst value is 0.
        scores["precision_score_micro"] = round(precision_score(test_true, test_preds, average="micro"), 5)
        scores["precision_score_macro"] = round(precision_score(test_true, test_preds, average="macro"), 5)
        scores["precision_score_weighted"] = round(precision_score(test_true, test_preds, average="weighted"), 5)

        # The recall is the ratio tp / (tp + fn) where tp is the number of true positives
        # and fn the number of false negatives. The recall is intuitively the ability
        # of the classifier to find all the positive samples.
        scores["recall_score_micro"] = round(recall_score(test_true, test_preds, average="micro"), 5)
        scores["recall_score_macro"] = round(recall_score(test_true, test_preds, average="macro"), 5)
        scores["recall_score_weighted"] = round(recall_score(test_true, test_preds, average="weighted"), 5)

        self.info("log_loss: %f", scores["log_loss"])
        self.info("accuracy_score: %f", scores["accuracy_score"])
        self.info("precision_score_micro: %f", scores["precision_score_micro"])
        self.info("precision_score_macro: %f", scores["precision_score_macro"])

        # complete classification report and confusion matrix
        # https://scikit-learn.org/stable/modules/generated/sklearn.metrics.classification_report.html#sklearn.metrics.classification_report
        # https://scikit-learn.org/stable/modules/generated/sklearn.metrics.confusion_matrix.html#sklearn.metrics.confusion_matrix
        scores["classification_report"] = classification_report(
            test_true, test_preds, target_names=train_classes, output_dict=True
        )
        scores["confusion_matrix"] = confusion_matrix(test_true, test_preds).tolist()

    def train(self, train, test, results, *args, **kwargs):
        """ Train with algorithm and given data to produce a trained model """
        try:
            assert isinstance(train, pd.DataFrame) and len(train.columns) > 1
            train_df = train
            test_df = test

            # if not specified the prediction target will be the last column of the dataset
            label = self.get_attribute("data.label")
            if not label:
                label = train_df.columns[len(train_df.columns) - 1]
            results["data"]["label"] = label

            # choose between regression, binary classification and multiclass classification
            label_type = analitico.schema.get_column_type(train_df, label)
            self.info("label: %s", label)
            self.info("label_type: %s", label_type)
            if label_type == analitico.schema.ANALITICO_TYPE_CATEGORY:
                label_classes = list(train_df[label].cat.categories)
                results["data"]["classes"] = label_classes
                train_df[label] = train_df[label].cat.codes
                results["algorithm"] = (
                    ALGORITHM_TYPE_BINARY_CLASSICATION
                    if len(label_classes) == 2
                    else ALGORITHM_TYPE_MULTICLASS_CLASSIFICATION
                )
                self.info("classes: %s", label_classes)
            else:
                results["algorithm"] = ALGORITHM_TYPE_REGRESSION
            self.info("algorithm: %s", results["algorithm"])

            # remove rows with missing label from training and test sets
            train_rows = len(train_df)
            train_df = train_df.dropna(subset=[label])
            if len(train_df) < train_rows:
                self.warning("Training data has %s rows without '%s' label", train_rows - len(train_df), label)
            if test_df:
                test_rows = len(test_df)
                test_df = test_df.dropna(subset=[label])
                if len(test_df) < test_rows:
                    self.warning("Test data has %s rows without '%s' label", test_rows - len(test_df), label)

            # make sure schemas match
            train_schema = self.validate_schema(train_df, test_df)

            # shortened training was requested?
            tail = self.get_attribute("parameters.tail", 0)
            if tail > 0:
                self.info("Tail: %d, cutting training data", tail)
                train_df = train_df.tail(tail).copy()

            # create test set from training set if not provided
            if not test_df:
                # decide how to create test set from settings variable
                # https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.TimeSeriesSplit.html
                chronological = self.get_attribute("data.chronological", False)
                test_size = self.get_attribute("parameters.test_size", 0.20)
                results["data"]["chronological"] = chronological
                results["parameters"]["test_size"] = test_size
                if chronological:
                    # test set if from the last rows (chronological order)
                    self.info("Test set split: chronological")
                    test_rows = int(len(train_df) * test_size)
                    test_df = train_df[-test_rows:]
                    train_df = train_df[:-test_rows]
                else:
                    # test set if from a random assortment of rows
                    self.info("Test set split: random")
                    train_df, test_df, = train_test_split(train_df, test_size=test_size, random_state=42)

            self.info("training: %d rows", len(train_df))
            self.info("testing: %d rows", len(test_df))

            # validate data types
            for column in train_schema["columns"]:
                if column["type"] not in ("integer", "float", "boolean", "category"):
                    self.warning(
                        "Column '%s' of type '%s' is incompatible and will be dropped", column["name"], column["type"]
                    )
                    train_df = train_df.drop(column["name"], axis=1)
                    test_df = test_df.drop(column["name"], axis=1)

            # save schema after dropping unused columns
            results["data"]["schema"] = generate_schema(train_df)
            results["data"]["source_records"] = len(train)
            results["data"]["training_records"] = len(train_df)
            results["data"]["test_records"] = len(test_df)
            results["data"]["dropped_records"] = len(train) - len(train_df) - len(test_df)

            # save some training data for debugging
            artifacts_path = self.factory.get_artifacts_directory()
            self.info("artifacts_path: %s", artifacts_path)

            samples_df = analitico.pandas.pd_sample(train_df, 200)
            samples_path = os.path.join(artifacts_path, "training-samples.json")
            samples_df.to_json(samples_path, orient="records")
            self.info("saved: %s (%d bytes)", samples_path, os.path.getsize(samples_path))
            samples_path = os.path.join(artifacts_path, "training-samples.csv")
            samples_df.to_csv(samples_path)
            self.info("saved: %s (%d bytes)", samples_path, os.path.getsize(samples_path))

            # split data and labels
            train_labels = train_df[label]
            train_df = train_df.drop([label], axis=1)
            test_labels = test_df[label]
            test_df = test_df.drop([label], axis=1)

            # indexes of columns that should be considered categorical
            categorical_idx = self.get_categorical_idx(train_df)
            train_pool = catboost.Pool(train_df, train_labels, cat_features=categorical_idx)
            test_pool = catboost.Pool(test_df, test_labels, cat_features=categorical_idx)

            # create regressor or classificator then train
            training_on = time_ms()
            model = self.create_model(results)
            model.fit(train_pool, eval_set=test_pool)
            results["performance"]["training_ms"] = time_ms(training_on)

            # score test set, add related metrics to results
            self.score_training(model, test_df, test_pool, test_labels, results)
            if results["algorithm"] == ALGORITHM_TYPE_REGRESSION:
                self.score_regressor_training(model, test_df, test_pool, test_labels, results)
            else:
                self.score_classifier_training(model, test_df, test_pool, test_labels, results)

            # save model file and training results
            model_path = os.path.join(artifacts_path, "model.cbm")
            model.save_model(model_path)
            results["scores"]["model_size"] = os.path.getsize(model_path)
            self.info("saved: %s (%d bytes)", model_path, os.path.getsize(model_path))
            return results

        except Exception as exc:
            self.exception("CatBoostPlugin - error while training: %s", str(exc), exception=exc)

    def predict(self, data, training, results, *args, **kwargs):
        """ Return predictions from trained model """

        # data should already come in as pd.DataFrame but it's just a dictionary we convert it
        if not isinstance(data, pd.DataFrame):
            data = pd.DataFrame.from_dict(data, orient="columns")

        # record that we're predicting on after augmentation is added
        # to the results. if the endpoint or the jupyter notebook in
        # charge of communicating with the caller does not want to send
        # this information back, it can always take it out. in the future
        # we may want to optimized here and add this optionally instead.
        results["records"] = analitico.pandas.pd_to_dict(data)

        # initialize data pool to be tested
        categorical_idx = self.get_categorical_idx(data)
        data_pool = catboost.Pool(data, cat_features=categorical_idx)

        # create model object from stored file
        loading_on = time_ms()
        model_path = os.path.join(self.factory.get_artifacts_directory(), "model.cbm")
        if not os.path.isfile(model_path):
            self.exception("CatBoostPlugin.predict - cannot find saved model in %s", model_path)

        model = self.create_model(training)
        model.load_model(model_path)
        results["performance"]["loading_ms"] = time_ms(loading_on)

        algo = training.get("algorithm", ALGORITHM_TYPE_REGRESSION)
        if algo == ALGORITHM_TYPE_REGRESSION:
            y_predictions = model.predict(data_pool)
            y_predictions = np.around(y_predictions, decimals=3)
            results["predictions"] = list(y_predictions)

        else:
            # predict class and probabilities of each class
            y_predictions = model.predict(
                data_pool, prediction_type="Class"
            )  # array di array of 1 element with class index
            y_probabilities = model.predict(data_pool, prediction_type="Probability")  # array of array of probabilities
            y_classes = training["data"]["classes"]  # list of possible classes

            preds = results["predictions"] = []
            probs = results["probabilities"] = []

            # create predictions with assigned class and probabilities
            if algo == ALGORITHM_TYPE_MULTICLASS_CLASSIFICATION:
                for i in range(0, len(data)):
                    preds.append(y_classes[int(y_predictions[i][0])])
                    probs.append({y_classes[j]: y_probabilities[i][j] for j in range(0, len(y_classes))})

            elif algo == ALGORITHM_TYPE_BINARY_CLASSICATION:
                for i in range(0, len(data)):
                    preds.append(y_classes[int(y_predictions[i])])
                    probs.append({y_classes[0]: y_probabilities[i][0], y_classes[1]: y_probabilities[i][1]})

        return results


##
## CatBoostRegressorPlugin
##


@plugin
class CatBoostRegressorPlugin(CatBoostPlugin):
    """ A tabular data regressor based on CatBoost library """

    class Meta(CatBoostPlugin.Meta):
        name = "analitico.plugin.CatBoostRegressorPlugin"
        algorithms = [ALGORITHM_TYPE_REGRESSION]


##
## CatBoostClassifierPlugin
##


@plugin
class CatBoostClassifierPlugin(CatBoostPlugin):
    """ A tabular data classifier based on CatBoost """

    class Meta(CatBoostPlugin.Meta):
        name = "analitico.plugin.CatBoostClassifierPlugin"
        algorithms = [ALGORITHM_TYPE_BINARY_CLASSICATION, ALGORITHM_TYPE_MULTICLASS_CLASSIFICATION]



import unittest
import os.path
import pandas as pd
import random
import string
import io
import json
import datetime

import analitico
from analitico import logger, authorize_sdk
from analitico.factory import Factory
from analitico.schema import generate_schema

from .test_mixin import TestMixin

# pylint: disable=no-member

ANALITICO_TEST_TOKEN = os.environ["ANALITICO_TEST_TOKEN"]
ANALITICO_TEST_WORKSPACE_ID = os.environ["ANALITICO_TEST_WORKSPACE_ID"]
assert ANALITICO_TEST_TOKEN, "The enviroment variable ANALITICO_TEST_TOKEN should be set with a test token."
assert (
    ANALITICO_TEST_WORKSPACE_ID
), "The enviroment variable ANALITICO_TEST_WORKSPACE_ID should be set with a test token."


TITANIC_PUBLIC_URL = "https://storage.googleapis.com/public.analitico.ai/data/titanic/train.csv"


class SDKTests(unittest.TestCase, TestMixin):
    """ Testing of Factory/sdk functionality: caching, creating items, plugins, etc """

    # Create a factory/sdk with out test token that can
    # access our test workspace in read and write model
    sdk = analitico.authorize_sdk(
        token=ANALITICO_TEST_TOKEN,
        workspace_id=ANALITICO_TEST_WORKSPACE_ID,
        endpoint="https://staging.analitico.ai/api/",
    )

    def setUp(self):
        logger.info("SDKTests.setUp")

    ##
    ## get
    ##

    def test_sdk_get_item(self):
        item = self.sdk.get_item("rx_ho374b88")
        self.assertIsNotNone(item)
        self.assertIsInstance(item, analitico.Recipe)
        self.assertTrue(item.id.startswith(analitico.RECIPE_PREFIX))
        self.assertEqual(item.type, "analitico/recipe")

    def test_sdk_get_recipe(self):
        recipe = self.sdk.get_recipe("rx_ho374b88")
        self.assertIsNotNone(recipe)
        self.assertIsInstance(recipe, analitico.Recipe)
        self.assertTrue(recipe.id.startswith(analitico.RECIPE_PREFIX))
        self.assertEqual(recipe.type, "analitico/recipe")

    ##
    ## create
    ##

    def test_sdk_create_item(self):
        item = None
        try:
            item = self.sdk.create_item(analitico.DATASET_TYPE)
            self.assertIsNotNone(item)
            self.assertIsInstance(item, analitico.Dataset)
            self.assertTrue(item.id.startswith(analitico.DATASET_PREFIX))
            self.assertEqual(item.type, "analitico/dataset")
        finally:
            if item:
                item.delete()

    def test_sdk_create_item_with_title(self):
        item = None
        try:
            title = "Testing at " + datetime.datetime.utcnow().isoformat()
            item = self.sdk.create_item(analitico.DATASET_TYPE, title=title)
            self.assertEqual(item.get_attribute("title"), title)
            self.assertEqual(item.title, title)
        finally:
            if item:
                item.delete()

    ##
    ## save
    ##

    def test_sdk_save_item_with_updates(self):
        item = None
        try:
            title = "Testing at " + datetime.datetime.utcnow().isoformat()
            item = self.sdk.create_item(analitico.DATASET_TYPE, title=title)
            self.assertEqual(item.get_attribute("title"), title)
            self.assertEqual(item.title, title)

            # update title, save updates on service
            title_v2 = title + " v2"
            item.title = title_v2
            item.save()
            self.assertEqual(item.get_attribute("title"), title_v2)
            self.assertEqual(item.title, title_v2)

            # retrieve new object from service
            item_again = self.sdk.get_item(item.id)
            self.assertEqual(item_again.get_attribute("title"), title_v2)
            self.assertEqual(item_again.title, title_v2)

        finally:
            if item:
                item.delete()

    ##
    ## upload
    ##

    def test_sdk_upload_dataframe(self):
        dataset = None
        try:
            title = "Boston test at " + datetime.datetime.utcnow().isoformat()
            dataset = self.sdk.create_item(analitico.DATASET_TYPE, title=title)

            from sklearn.datasets import load_boston

            boston_dataset = load_boston()
            boston_df = pd.DataFrame(boston_dataset.data, columns=boston_dataset.feature_names)
            dataset.upload(df=boston_df)

        finally:
            if dataset:
                dataset.delete()


    def test_upload_ilpes(self):
        sdk = analitico.authorize_sdk(endpoint="https://staging.analitico.ai/api/", token="tok_pbn6gxbj", workspace_id="ws_2q2fq3wg")

        dataset = sdk.get_dataset("ds_aqp195k4")

        from sklearn.datasets import load_boston
        boston_dataset = load_boston()
        boston_df = pd.DataFrame(boston_dataset.data, columns=boston_dataset.feature_names)
        dataset.upload("boston.parquet", df=boston_df)




import pandas as pd

import analitico.pandas
import analitico.utilities

from analitico.constants import ACTION_TRAIN
from analitico.utilities import time_ms, timeit, get_dict_dot

from .interfaces import IDataframeSourcePlugin, plugin

##
## DatasetSourcePlugin
##


@plugin
class DatasetSourcePlugin(IDataframeSourcePlugin):
    """ A plugin that returns data from an analitico dataset """

    class Meta(IDataframeSourcePlugin.Meta):
        name = "analitico.plugin.DatasetSourcePlugin"

    @timeit
    def retrieve_df(self, *args, action=None, **kwargs):
        """ Retrieve dataframe from dataset with id set in plugin's configuration """
        try:
            dataset_id = self.get_attribute("dataset_id")
            if not dataset_id:
                dataset_id = self.get_attribute("source.dataset_id")
                if not dataset_id:
                    self.exception("DatasetSourcePlugin - must specify 'dataset_id'")

            info_url = "analitico://datasets/" + dataset_id + "/data/info"
            self.info("reading: %s", info_url)

            info = self.factory.get_url_json(info_url)
            schema = get_dict_dot(info, "data.schema", None)
            if not schema:
                self.warning("DatasetSourcePlugin - %s does not contain schema information", info_url)

            # save the schema for the source so it can be used to enforce it on prediction
            self.set_attribute("source.schema", schema)

            # stream data from dataset endpoint or storage as csv
            csv_url = "analitico://datasets/" + dataset_id + "/data/csv"
            csv_stream = self.factory.get_url_stream(csv_url, binary=False)

            reading_on = time_ms()
            self.info("reading: %s", csv_url)
            df = analitico.pandas.pd_read_csv(csv_stream, schema)
            self.info("%d rows in %d ms", len(df), time_ms(reading_on))

            # https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.sample.html
            sample = self.get_attribute("sample", 0)
            if sample > 0:
                rows_before = len(df)
                df = analitico.pandas.pd_sample(df, sample)
                self.info("sample: %f, rows before: %d, rows after: %d", sample, rows_before, len(df))

            tail = self.get_attribute("tail", 0)
            if tail > 0:
                rows_before = len(df)
                df = df.tail(tail)
                self.info("tail: %d, rows before: %d, rows after: %d", tail, rows_before, len(df))

            return df

        except Exception as exc:
            raise exc

    def run(self, *args, action=None, **kwargs):
        """ Read data from configured dataset in training mode, noop in prediction mode """

        # DEPRECATED in training one may leave this plugin in but pass a dataframe for inference
        # so that the same pipeline can be reused for training and prediction. in this case we will
        # do nothing and pass data through
        if len(args) > 0 and isinstance(args[0], pd.DataFrame):
            self.factory.warning("DatasetSourcePlugin - passed a DataFrame as input, will noop")
            return args[0]

        return self.retrieve_df(*args, action, **kwargs)



import os
import os.path
import pandas as pd

import analitico.pandas
from analitico.utilities import read_json, get_dict_dot

from .interfaces import plugin
from .pipelineplugin import PipelinePlugin

##
## EndpointPipelinePlugin
##


@plugin
class EndpointPipelinePlugin(PipelinePlugin):
    """
    EndpointPipelinePlugin is a base class for endpoints that take trained machine
    learning models to deliver inferences. An endpoint subclass could implement 
    inference APIs by taking a web request and returning predictions, etc.
    """

    class Meta(PipelinePlugin.Meta):
        name = "analitico.plugin.EndpointPipelinePlugin"
        inputs = [{"data": "pandas.DataFrame"}]
        outputs = [{"predictions": "pandas.DataFrame"}]

    def run(self, *args, action=None, **kwargs):
        """ Process the plugins in sequence to run predictions """
        try:
            # if no plugins have been configured for the pipeline,
            # create the plugin suggested by the training algorithm
            if not self.plugins:
                # read training information from disk
                artifacts_path = self.factory.get_artifacts_directory()
                training_path = os.path.join(artifacts_path, "training.json")
                training = read_json(training_path)
                assert training
                self.set_attribute("plugins", [{"name": get_dict_dot(training, "plugins.prediction")}])

            assert isinstance(args[0], pd.DataFrame)
            df = args[0]
            df_copy = df.copy()

            # run the pipeline, return predictions
            predictions = super().run(df, action=action, **kwargs)

            # the predictor has most likely added a "records" field with the processed
            # records which we may choose to avoid echoing back to the caller. if we want
            # we can remove the records here before returning

            return predictions

        except Exception as exc:
            self.error("Error while processing prediction pipeline")
            self.logger.exception(exc)
            raise exc



from .test_pandas import PandasTests
from .test_utilities import UtilitiesTests
from .test_catboost import CatBoostTests
from .test_dataset import DatasetTests
from .test_factory import FactoryTests
from .test_mixin import TestMixin
from .test_plugin import PluginTests
from .test_metadata import MetadataTests
from .test_sdk import SDKTests



import unittest
import tempfile
import numpy as np
import pytest

import pandas as pd
import numpy as np
from datetime import datetime, timedelta

from analitico.pandas import *


@pytest.mark.django_db
class PandasTests(unittest.TestCase):
    def get_random_dates_df(self):
        date_today = datetime.now()
        dates1 = pd.date_range(date_today, date_today + timedelta(21), freq="H")
        dates2 = pd.date_range(date_today, date_today + timedelta(21), freq="D")

        np.random.seed(seed=1111)
        data1 = np.random.randint(1, high=100, size=len(dates1))
        data2 = np.random.randint(1, high=100, size=len(dates1))
        data3 = np.random.randint(1, high=100, size=len(dates1))

        return pd.DataFrame({"Data1": data1, "Dates1": dates1, "Dates2": dates1, "Data2": data2, "Data3": data3})

    def test_pandas_augment_dates_all_columns(self):
        df1 = self.get_random_dates_df()
        df2 = augment_dates(df1)

        self.assertIn("Dates1.dayofweek", df2.columns)
        self.assertIn("Dates1.year", df2.columns)
        self.assertIn("Dates1.month", df2.columns)
        self.assertIn("Dates1.day", df2.columns)
        self.assertIn("Dates1.hour", df2.columns)
        self.assertIn("Dates1.minute", df2.columns)
        self.assertNotIn("Dates1", df2.columns)

        self.assertIn("Dates2.dayofweek", df2.columns)
        self.assertIn("Dates2.year", df2.columns)
        self.assertIn("Dates2.month", df2.columns)
        self.assertIn("Dates2.day", df2.columns)
        self.assertIn("Dates2.hour", df2.columns)
        self.assertIn("Dates2.minute", df2.columns)
        self.assertNotIn("Dates2", df2.columns)

    def test_pandas_augment_dates_only_dates1(self):
        df1 = self.get_random_dates_df()
        df2 = augment_dates(df1, column="Dates1")

        self.assertIn("Dates1.dayofweek", df2.columns)
        self.assertIn("Dates1.year", df2.columns)
        self.assertIn("Dates1.month", df2.columns)
        self.assertIn("Dates1.day", df2.columns)
        self.assertIn("Dates1.hour", df2.columns)
        self.assertIn("Dates1.minute", df2.columns)

        self.assertNotIn("Dates1", df2.columns)
        self.assertIn("Dates2", df2.columns)

    def test_pandas_augment_dates_only_dates2(self):
        df1 = self.get_random_dates_df()
        df2 = augment_dates(df1, column="Dates2")

        self.assertIn("Dates2.dayofweek", df2.columns)
        self.assertIn("Dates2.year", df2.columns)
        self.assertIn("Dates2.month", df2.columns)
        self.assertIn("Dates2.day", df2.columns)
        self.assertIn("Dates2.hour", df2.columns)
        self.assertIn("Dates2.minute", df2.columns)

        self.assertNotIn("Dates2", df2.columns)
        self.assertIn("Dates1", df2.columns)

    def test_pandas_augment_dates_only_dayofweek(self):
        df1 = self.get_random_dates_df()
        df2 = augment_dates(df1, column="Dates2", expand=["dayofweek"])

        self.assertIn("Dates2.dayofweek", df2.columns)
        self.assertNotIn("Dates2.year", df2.columns)
        self.assertNotIn("Dates2.month", df2.columns)
        self.assertNotIn("Dates2.day", df2.columns)
        self.assertNotIn("Dates2.hour", df2.columns)
        self.assertNotIn("Dates2.minute", df2.columns)

        self.assertNotIn("Dates2", df2.columns)
        self.assertIn("Dates1", df2.columns)

    def test_pandas_augment_dates_only_dayofweek_and_year(self):
        df1 = self.get_random_dates_df()
        df2 = augment_dates(df1, column="Dates2", expand=["dayofweek", "year"])

        self.assertIn("Dates2.dayofweek", df2.columns)
        self.assertIn("Dates2.year", df2.columns)
        self.assertNotIn("Dates2.month", df2.columns)
        self.assertNotIn("Dates2.day", df2.columns)
        self.assertNotIn("Dates2.hour", df2.columns)
        self.assertNotIn("Dates2.minute", df2.columns)

        self.assertNotIn("Dates2", df2.columns)
        self.assertIn("Dates1", df2.columns)

    def test_pandas_augment_dates_check_position_2(self):
        df1 = self.get_random_dates_df()
        df2 = augment_dates(df1, column="Dates2", expand=["dayofweek", "year"])

        cols = list(df2.columns)
        self.assertEqual(cols.index("Dates2.dayofweek"), 2)
        self.assertEqual(cols.index("Dates2.year"), 3)
        self.assertEqual(cols.index("Dates1"), 1)

    def test_pandas_augment_dates_check_position_1(self):
        df1 = self.get_random_dates_df()
        df2 = augment_dates(df1, column="Dates1", expand=["dayofweek", "year"])

        cols = list(df2.columns)
        self.assertEqual(cols.index("Dates1.dayofweek"), 1)
        self.assertEqual(cols.index("Dates1.year"), 2)
        self.assertEqual(cols.index("Dates2"), 3)

    def test_pandas_augment_dates_check_position_no_drop(self):
        df1 = self.get_random_dates_df()
        df2 = augment_dates(df1, column="Dates1", expand=["dayofweek", "year"], drop=False)

        cols = list(df2.columns)
        self.assertIn("Dates1", cols)
        self.assertEqual(cols.index("Dates1"), 1)
        self.assertEqual(cols.index("Dates1.dayofweek"), 2)
        self.assertEqual(cols.index("Dates1.year"), 3)
        self.assertEqual(cols.index("Dates2"), 4)

    def test_pandas_augment_dates_check_position_no_drop(self):
        df1 = self.get_random_dates_df()
        df1["Dates1"] = df1["Dates1"].astype("object")  # convert to strings
        self.assertEqual(df1["Dates1"].dtype, "object")

        df2 = augment_dates(df1, column="Dates1", expand=["dayofweek", "year"], drop=False)

        cols = list(df2.columns)
        self.assertIn("Dates1", cols)
        self.assertEqual(df1["Dates1"].dtype, "datetime64[ns]")
        self.assertNotEqual(df1["Dates1"].dtype, "object")
        self.assertEqual(cols.index("Dates1"), 1)
        self.assertEqual(cols.index("Dates1.dayofweek"), 2)
        self.assertEqual(cols.index("Dates1.year"), 3)
        self.assertEqual(cols.index("Dates2"), 4)

    def test_pandas_augment_dates_check_position_check_values(self):
        df1 = self.get_random_dates_df()
        df2 = augment_dates(df1, column="Dates1", drop=False)

        # vector comparison
        dates = pd.DatetimeIndex(df2["Dates1"])
        self.assertTrue((dates.dayofweek == df2["Dates1.dayofweek"]).all())
        self.assertTrue((dates.year == df2["Dates1.year"]).all())
        self.assertTrue((dates.month == df2["Dates1.month"]).all())
        self.assertTrue((dates.day == df2["Dates1.day"]).all())
        self.assertTrue((dates.hour == df2["Dates1.hour"]).all())
        self.assertTrue((dates.minute == df2["Dates1.minute"]).all())

    def test_pandas_augment_dates_check_position_check_categoricals(self):
        df1 = self.get_random_dates_df()
        df2 = augment_dates(df1, column="Dates1")

        self.assertEqual(df2["Dates1.dayofweek"].dtype, "category")
        self.assertEqual(df2["Dates1.year"].dtype, "category")
        self.assertEqual(df2["Dates1.month"].dtype, "category")
        self.assertEqual(df2["Dates1.day"].dtype, "category")
        self.assertEqual(df2["Dates1.hour"].dtype, "category")
        self.assertEqual(df2["Dates1.minute"].dtype, "category")



import pandas as pd

from analitico.schema import apply_schema
from .interfaces import IDataframePlugin, plugin

##
## TransformDataframePlugin
##


@plugin
class TransformDataframePlugin(IDataframePlugin):
    """
    TransformDataframePlugin can apply a schema to a dataframe and can be used to:
    - drop columns you don't need
    - apply a type to a column (eg. change a string to a date)
    - reorder columns in a dataframe (eg. put the label last)
    - rename columns
    - make a column the index of the dataframe
    """

    class Meta(IDataframePlugin.Meta):
        name = "analitico.plugin.TransformDataframePlugin"
        title = "TransformDataframePlugin"
        description = "This plugin applies a schema to the input dataframe to provide a variety of transformations."
        configurations = [
            {
                "name": "schema",
                "type": "analitico/schema",
                "optional": False,
                "description": "The schema is applied to the input dataframe. The choice and attributes "
                + "of columns indicate how they are filtered, ordered, cast, etc.",
            }
        ]

    def run(self, *args, action=None, **kwargs) -> pd.DataFrame:

        df = args[0]
        if not isinstance(df, pd.DataFrame):
            self.warning("TransformDataframePlugin - requires a single pd.DataFrame as input, none was found")
            return args

        schema = self.get_attribute("schema", None)
        if not schema:
            self.warning(
                "TransformDataframePlugin - should have a 'schema' attribute with the schema that you want to apply to the input dataframe"
            )
            return df

        df = apply_schema(df, schema)
        self.info("TransformDataframePlugin - schema applied to dataframe")

        return df



import tempfile
import os
import os.path
import requests
import json
import re
import logging
import hashlib
import inspect
import urllib.parse
import io
import pandas as pd
import tempfile

from .mixin import AttributeMixin
from .exceptions import AnaliticoException
from .status import STATUS_FAILED

import analitico.utilities
import analitico.models

from analitico.utilities import id_generator
from analitico.models import Workspace, Item, Dataset, Recipe, Notebook

import logging

logger = logging.getLogger("analitico")

# read http streams in chunks
HTTP_BUFFER_SIZE = 32 * 1024 * 1024  # 32 MiBs

##
## Models used in the SDK
##


class AnaliticoSDK(AttributeMixin):
    """ An SDK for analitico.ai/api. """

    def __init__(self, token=None, endpoint=None, workspace_id: str = None, **kwargs):
        super().__init__(**kwargs)
        if token:
            assert token.startswith("tok_")
            self.set_attribute("token", token)
        if endpoint:
            assert endpoint.startswith("http")
            self.set_attribute("endpoint", endpoint)

        # set default workspace
        if workspace_id:
            self.workspace = self.get_workspace(workspace_id)

        # use current working directory at the time when the factory
        # is created so that the caller can setup a temp directory we
        # should work in
        self._artifacts_directory = os.getcwd()

    ##
    ## Properties and factory context
    ##

    # default workspace
    _workspace: Workspace = None

    @property
    def workspaces(self) -> [Workspace]:
        """ Returns your workspaces. """
        return self.get_items(analitico.WORKSPACE_TYPE)

    @property
    def workspace(self):
        """ Get the default workspace used by the SDK (for example when creating items). """
        return self._workspace

    @workspace.setter
    def workspace(self, workspace: Workspace):
        self._workspace = workspace

    @property
    def token(self):
        """ API token used to call endpoint (optional) """
        return self.get_attribute("token")

    @property
    def endpoint(self):
        """ Endpoint used to call analitico APIs """
        return self.get_attribute("endpoint")

    ##
    ## Temp and cache directories
    ##

    # Temporary directory which is deleted when factory is disposed
    _temp_directory = None

    # Artifacts end up in the current working directory
    _artifacts_directory = None

    def get_temporary_directory(self):
        """ Temporary directory that can be used while a factory is used and deleted afterwards """
        temp_dir = os.path.join(tempfile.gettempdir(), "analitico_temp")
        if not os.path.isdir(temp_dir):
            os.mkdir(temp_dir)
        return temp_dir

    def get_artifacts_directory(self):
        """ 
        A plugin or notebook can produce various file artifacts during execution and place
        them in this directory (datasets, statistics, models, etc). A subclass, for example
        a factory used to run pipelines on the server, may persist files created here to cloud, etc.
        """
        return self._artifacts_directory

    def get_cache_directory(self):
        """ Returns directory to be used for caches """
        cache_dir = os.path.join(tempfile.gettempdir(), "analitico_cache")
        if not os.path.isdir(cache_dir):
            os.mkdir(cache_dir)
        return cache_dir

    def get_cache_filename(self, unique_id):
        """ Returns the fullpath in cache for an item with the given unique_id (eg: a unique url, an md5 or etag, etc) """
        # Tip: if cache contents need to be invalidated for whatever reason, you can change the prefix below...
        return os.path.join(self.get_cache_directory(), "cache_v2_" + hashlib.sha256(unique_id.encode()).hexdigest())

    ##
    ## Internal Utilities
    ##

    # regular expression used to detect assets using analitico:// scheme
    ANALITICO_ASSET_RE = r"(analitico://workspaces/(?P<workspace_id>[-\w.]{4,256})/)"

    EMAIL_RE = r"[^@]+@[^@]+\.[^@]+"  # very rough check

    def get_item_type(self, item_id):
        """ Returns item class from item id, eg: returns 'dataset' from ds_xxx """
        assert item_id and isinstance(item_id, str)
        if item_id.startswith(analitico.DATASET_PREFIX):
            return analitico.DATASET_TYPE
        if item_id.startswith(analitico.ENDPOINT_PREFIX):
            return analitico.ENDPOINT_TYPE
        if item_id.startswith(analitico.JOB_PREFIX):
            return analitico.JOB_TYPE
        if item_id.startswith(analitico.MODEL_PREFIX):
            return analitico.MODEL_TYPE
        if item_id.startswith(analitico.NOTEBOOK_PREFIX):
            return analitico.NOTEBOOK_TYPE
        if item_id.startswith(analitico.PLUGIN_PREFIX):
            return analitico.PLUGIN_TYPE
        if item_id.startswith(analitico.RECIPE_PREFIX):
            return analitico.RECIPE_TYPE
        if item_id.startswith(analitico.WORKER_PREFIX):
            return analitico.WORKER_TYPE
        if item_id.startswith(analitico.WORKSPACE_PREFIX):
            return analitico.WORKSPACE_TYPE
        if re.match(self.EMAIL_RE, item_id):
            return analitico.USER_TYPE
        self.warning("Factory.get_item_type - couldn't find type for: " + item_id)
        return None

    # TODO could check if it would be possible to switch to an external caching library, eg:
    # https://github.com/ionrock/cachecontrol

    def get_cached_stream(self, stream, unique_id):
        """ Will cache a stream on disk based on a unique_id (like md5 or etag) and return file stream and filename """
        cache_file = self.get_cache_filename(unique_id)
        if not os.path.isfile(cache_file):
            # if not cached already, download and cache
            cache_temp_file = cache_file + ".tmp_" + id_generator()

            with open(cache_temp_file, "wb") as f:
                if hasattr(stream, "read"):
                    for chunk in iter(lambda: stream.read(HTTP_BUFFER_SIZE), b""):
                        f.write(chunk)
                else:
                    for b in stream:
                        f.write(b)

            # TODO add progress bar for slow downloads https://github.com/tqdm/tqdm#iterable-based

            os.rename(cache_temp_file, cache_file)
        # return stream from cached file
        return open(cache_file, "rb"), cache_file

    def get_url_headers(self, url: str) -> (str, dict):
        # If the url uses the analitico:// scheme for assets stored on the cloud
        # service, it will convert the url to a regular https:// scheme.
        # If the url points to an analitico API call, the request will have the
        # ?token= authorization token header added to it.
        # temporarily while all internal urls are updated to analitico://
        assert url and isinstance(url, str)
        if url.startswith("workspaces/ws_"):
            url = analitico.ANALITICO_URL_PREFIX + url
        # see if assets uses analitico://workspaces/... scheme
        if url.startswith("analitico://"):
            if not self.endpoint:
                raise AnaliticoException(
                    f"Analitico SDK is not configured with a valid API endpoint and cannot get {url}"
                )
            url = self.endpoint + url[len("analitico://") :]

        try:
            url_parse = urllib.parse.urlparse(url)
        except Exception:
            pass
        headers = {}
        if url_parse and url_parse.scheme in ("http", "https"):
            if url_parse.hostname and url_parse.hostname.endswith("analitico.ai") and self.token:
                # if url is connecting to analitico.ai add token
                headers = {"Authorization": "Bearer " + self.token}
        return url, headers

    def get_url_stream(
        self,
        url: str,
        data: dict = None,
        json: dict = None,
        files: dict = None,
        binary: bool = False,
        cache: bool = True,
        method: str = "GET",
        status_code: int = 200,
    ):
        """
        Returns a stream to the given url. This works for regular http:// or https://
        and also works for analitico:// assets which are converted to calls to the given
        endpoint with proper authorization tokens. The stream is returned as an iterator.
        """
        url, headers = self.get_url_headers(url)
        # we should not take the raw response stream here as it could be gzipped or encoded.
        # we take the decoded content as a text string and turn it into a stream or we take the
        # decompressed binary content and also turn it into a stream.
        response = requests.request(method, url, data=data, files=files, stream=True, headers=headers)
        if status_code and response.status_code != status_code:
            msg = f"The response from {url} should have been {status_code} but instead it is {response.status_code}."
            raise AnaliticoException(msg)
        # always treat content as binary, utf-8 encoding is done by readers
        response_stream = io.BytesIO(response.content)
        return response_stream

    def get_url_json(self, url: str, json: dict = None, method: str = "GET", status_code: int = 200) -> dict:
        """
        Get a json response from given url. If the url starts with analitico:// it will be
        substituted with the url of the actual endpoint for analitico.ai and a bearer token
        will be attached for authorization.
        
        Arguments:
            url {str} -- An absolute url to be read or an analitico:// url for API calls.
            data {dict} -- An optional dictionary that should be sent (eg: for POST calls).
        
        Keyword Arguments:
            method {str} -- HTTP method to be used (default: {"get"})
        
        Returns:
            dict -- The json response.
        """
        url, headers = self.get_url_headers(url)

        response = requests.request(method, url, headers=headers, json=json)
        if status_code and response.status_code != status_code:
            msg = f"The response from {url} should have been {status_code} but instead it is {response.status_code}."
            raise AnaliticoException(msg)
        try:
            return response.json()
        except Exception:
            return None

    ##
    ## with Factory as: lifecycle methods
    ##

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        """ Leave any temporary files upon exiting """
        pass

    ##
    ## SDK v1 methods
    ##

    def create_item(self, item_type: str, workspace: Workspace = None, **kwargs) -> Item:
        if not workspace:
            workspace = self.get_workspace()

        data = {"data": kwargs}
        data["data"]["workspace_id"] = workspace.id
        if "type" in kwargs:
            kwargs.pop("type")
        if "id" in kwargs:
            data["id"] = kwargs.pop("id")

        json = self.get_url_json(f"analitico://{item_type}s", json=data, method="POST", status_code=201)
        assert "data" in json
        return analitico.models.models_factory(self, json["data"])

    def create_dataset(self, workspace: Workspace = None, **kwargs) -> Dataset:
        return self.create_item(workspace=workspace, item_type=analitico.DATASET_TYPE, **kwargs)

    def create_recipe(self, workspace: Workspace = None, **kwargs) -> Recipe:
        return self.create_item(workspace=workspace, item_type=analitico.RECIPE_TYPE, **kwargs)

    def create_notebook(self, workspace: Workspace = None, **kwargs) -> Notebook:
        return self.create_item(workspace=workspace, item_type=analitico.NOTEBOOK_TYPE, **kwargs)

    ##
    ## Retrieve specific items by id
    ##

    def get_items(self, item_type: str) -> Item:
        """ Retrieves item from the server by item_id """
        json = self.get_url_json(f"analitico://{item_type}s")
        items = []
        for item_data in json["data"]:
            items.append(analitico.models.models_factory(self, item_data))
        return items

    def get_item(self, item_id: str) -> Item:
        """ Retrieves item from the server by item_id """
        json = self.get_url_json(f"analitico://{self.get_item_type(item_id)}s/{item_id}")
        assert "data" in json
        return analitico.models.models_factory(self, json["data"])

    def get_workspace(self, workspace_id: str = None) -> Workspace:
        if not workspace_id:
            if not self.workspace:
                workspaces = self.get_items(analitico.WORKSPACE_TYPE)
                if len(workspaces) != 1:
                    raise AnaliticoException("You do not have a default workspace, please assign sdk.workspace")
                self.workspace = workspaces[0]
            return self.workspace

        workspace = self.get_item(workspace_id)
        assert isinstance(workspace, Workspace)
        return workspace

    def get_dataset(self, dataset_id) -> Dataset:
        dataset = self.get_item(dataset_id)
        assert isinstance(dataset, Dataset)
        return dataset

    def get_recipe(self, recipe_id) -> Recipe:
        recipe = self.get_item(recipe_id)
        assert isinstance(recipe, Recipe)
        return recipe

    def get_notebook(self, notebook_id) -> Notebook:
        notebook = self.get_item(notebook_id)
        assert isinstance(notebook, Notebook)
        return notebook



import os
import json
import logging
import collections

import sklearn
import sklearn.metrics

from analitico import logger
from analitico.utilities import read_json, save_json, get_dict_dot, set_dict_dot

# model metadata is saved in training.json
METADATA_FILENAME = "training.json"


def get_metadata(metadata_filename=METADATA_FILENAME):
    """ Returns metadata dictionary for current recipe. """
    return read_json(metadata_filename) if os.path.isfile(metadata_filename) else collections.OrderedDict()


def set_metric(
    metric: str,
    value,
    title: str = None,
    subtitle: str = None,
    priority: int = None,
    category: str = None,
    category_title: str = None,
    category_subtitle: str = None,
    metadata_filename=METADATA_FILENAME,
):
    """
    Collect a specific metric in the metadata file.

    These metrics can be useful to track performance of a trained model and can 
    also be shown in analitico's UI next to the trained models' information. A score
    has a machine readable id like number_of_lines and a value, eg. 100.

    Parameters:
    metric (str): The machine readable name of the metric, eg: number_of_records.
    value (object): The value of the metric can be an integer, float, string or a dictionary.
    title (str): Human readable name of the metric, eg: Number of records (optional).
    subtitle (str): Human readable name of the metric, eg: Number of records (optional).
    priority (int): Metric priority when shown in the UI (eg. 1 higher, 10 lower) (optional).
    category (str): A category id that should be used to group these metrics, eg. svc_classifier (optional).
    category_title (str): Human readable title for the category (optional).
    category_subtitle (str): A subtitle or tooltip for the category with more details (optional).
    metadata_filename (str): Path of metadata filename to be used if not the default one (optional).

    Returns:
    Nothing
    """
    metadata = get_metadata()

    if title or priority:
        value = {"value": value}
        if title:
            value["title"] = title
        if subtitle:
            value["subtitle"] = subtitle
        if priority:
            value["priority"] = priority

    key = f"scores.{category}.{metric}" if category else f"scores.{metric}"
    set_dict_dot(metadata, key, value)

    if category:
        if category_title:
            set_dict_dot(metadata, f"scores.{category}.title", category_title)
        if category_subtitle:
            set_dict_dot(metadata, f"scores.{category}.subtitle", category_subtitle)

    save_json(metadata, METADATA_FILENAME)


def set_model_metrics(
    y_true,
    y_pred,
    model: object = None,
    is_regressor: bool = None,
    is_classifier: bool = None,
    target_names=None,
    category: str = None,
    category_title: str = None,
    category_subtitle: str = None,
    metadata_filename: str = METADATA_FILENAME,
):
    """
    Save metrics related to a model's performance.

    Takes a model (derived from sklearn base estimator) and two array of values
    and predictions and saves a number of statistical scores regarding the accuracy
    of the predictions.

    Parameters:
    y_true (1d array-like): Ground truth (correct) labels/values.
    y_pred (1d array-line): Predicted labels or values, as returned by the model.
    model (object): An sklearn model that was used for predicting (optional).
    is_regressor (bool): True if model is a regressor, can also be inferred from model (optional).
    is_classifier (bool): True if model is a classifier, can also be inferred from model (optional).
    target_names (array): An array of labels for classifiers (optional).
    category (str): A category id that should be used to group these metrics, eg. svc_classifier (optional).
    category_title (str): Human readable title for the category (optional).
    category_subtitle (str): A subtitle or tooltip for the category with more details (optional).
    metadata_filename (str): Path of metadata filename to be used if not the default one (optional).

    Returns:
    Nothing
    """
    extras = {
        "category": category if category else "sklearn_metrics",
        "category_title": category_title if category_title else "Scikit Learn Metrics",
        "category_subtitle": category_subtitle if category_subtitle else None,
    }

    # ask sklearn what kind of model we have
    if isinstance(model, sklearn.base.BaseEstimator):
        is_regressor = sklearn.base.is_regressor(model) if is_regressor is None else is_regressor
        is_classifier = sklearn.base.is_classifier(model) if is_classifier is None else is_classifier
    else:
        # TODO we could probably guess the kind of model by just looking at the data
        logger.warning("set_model_metrics - we only support sklearn models for now")

    # There are many metrics for regression, we pick some basic ones...
    # https://scikit-learn.org/stable/modules/model_evaluation.html#regression-metrics
    if is_regressor:
        set_metric(
            metric="mean_abs_error",
            value=round(sklearn.metrics.mean_absolute_error(y_true, y_pred), 5),
            title="Mean absolute regression loss",
            subtitle="https://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_absolute_error.html",
            **extras,
        )
        set_metric(
            metric="median_abs_error",
            value=round(sklearn.metrics.median_absolute_error(y_true, y_pred), 5),
            title="Median absolute error regression loss",
            subtitle="https://scikit-learn.org/stable/modules/generated/sklearn.metrics.median_absolute_error.html",
            **extras,
        )
        set_metric(
            metric="mean_squared_error",
            value=round(sklearn.metrics.mean_squared_error(y_true, y_pred), 5),
            title="Mean squared error regression loss",
            subtitle="https://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_squared_error.html",
            **extras,
        )

    # https://scikit-learn.org/stable/modules/model_evaluation.html#classification-metrics
    if is_classifier:
        classification_report = sklearn.metrics.classification_report(
            y_true, y_pred, target_names=target_names, output_dict=True
        )
        set_metric(
            metric="classification_report",
            value=classification_report,
            title="Classification report",
            subtitle="https://scikit-learn.org/stable/modules/generated/sklearn.metrics.classification_report.html",
            **extras,
        )
        # TODO could use target_names?
        confusion_matrix = sklearn.metrics.confusion_matrix(y_true, y_pred).tolist()
        set_metric(
            metric="confusion_matrix",
            value=confusion_matrix,
            title="Confusion matrix",
            subtitle="https://scikit-learn.org/stable/modules/generated/sklearn.metrics.confusion_matrix.html",
            **extras,
        )

        # classification report has 3 fixed keys plus one per class
        is_binary_classifier = (len(classification_report) - 3) == 2

        if is_binary_classifier:
            set_metric(
                metric="log_loss",
                value=round(sklearn.metrics.log_loss(y_true, y_pred), 5),
                title="Log loss, aka logistic loss or cross-entropy loss",
                subtitle="https://scikit-learn.org/stable/modules/generated/sklearn.metrics.log_loss.html",
                **extras,
            )

        # In multilabel classification, this function computes subset accuracy:
        # the set of labels predicted for a sample must exactly match the corresponding set of labels in y_true.
        set_metric(
            metric="accuracy_score",
            value=round(sklearn.metrics.accuracy_score(y_true, y_pred), 5),
            title="Accuracy classification score",
            subtitle="https://scikit-learn.org/stable/modules/generated/sklearn.metrics.accuracy_score.html",
            **extras,
        )

        # The precision is the ratio tp / (tp + fp) where tp is the number of true positives
        # and fp the number of false positives. The precision is intuitively the ability
        # of the classifier not to label as positive a sample that is negative.
        # The best value is 1 and the worst value is 0.
        set_metric(
            metric="precision_score_micro",
            value=round(sklearn.metrics.precision_score(y_true, y_pred, average="micro"), 5),
            title="Precision score (micro)",
            subtitle="https://scikit-learn.org/stable/modules/generated/sklearn.metrics.precision_score.html",
            **extras,
        )
        set_metric(
            metric="precision_score_macro",
            value=round(sklearn.metrics.precision_score(y_true, y_pred, average="macro"), 5),
            title="Precision score (macro)",
            subtitle="https://scikit-learn.org/stable/modules/generated/sklearn.metrics.precision_score.html",
            **extras,
        )
        set_metric(
            metric="precision_score_weighted",
            value=round(sklearn.metrics.precision_score(y_true, y_pred, average="weighted"), 5),
            title="Precision score (weighted)",
            subtitle="https://scikit-learn.org/stable/modules/generated/sklearn.metrics.precision_score.html",
            **extras,
        )

        # The recall is the ratio tp / (tp + fn) where tp is the number of true positives
        # and fn the number of false negatives. The recall is intuitively the ability
        # of the classifier to find all the positive samples.
        set_metric(
            metric="recall_score_micro",
            value=round(sklearn.metrics.recall_score(y_true, y_pred, average="micro"), 5),
            title="Recall score (micro)",
            subtitle="https://scikit-learn.org/stable/modules/generated/sklearn.metrics.recall_score.html",
            **extras,
        )
        set_metric(
            metric="recall_score_macro",
            value=round(sklearn.metrics.recall_score(y_true, y_pred, average="macro"), 5),
            title="Recall score (macro)",
            subtitle="https://scikit-learn.org/stable/modules/generated/sklearn.metrics.recall_score.html",
            **extras,
        )
        set_metric(
            metric="recall_score_weighted",
            value=round(sklearn.metrics.recall_score(y_true, y_pred, average="weighted"), 5),
            title="Recall score (weighted)",
            subtitle="https://scikit-learn.org/stable/modules/generated/sklearn.metrics.recall_score.html",
            **extras,
        )



# schemas
ANALITICO_SCHEMA = "analitico"
ANALITICO_URL_PREFIX = "analitico://"

# API endpoints
ANALITICO_STAGING_API_ENDPOINT = "https://staging.analitico.ai/api/"
ANALITICO_API_ENDPOINT = "https://analitico.ai/api/"

# actions
ACTION_PROCESS = "process"  # process a dataframe to retrieve data
ACTION_TRAIN = "train"  # train a recipe to procude a model
ACTION_PREDICT = "predict"  # run a model to generate a prediction
ACTION_DEPLOY = "deploy"  # deploy a model to a serverless endpoint

ACTION_RUN = "run"  # run a recipe, notebook or dataset notebooks
ACTION_BUILD = "build"  # build a snapshot of a recipe into a docker
ACTION_RUN_AND_BUILD = "run-and-build"  # run the recipe then build its docker/model as one job with two steps

# types/models
TYPE_PREFIX = "analitico/"
DATASET_TYPE = "dataset"
ENDPOINT_TYPE = "endpoint"
JOB_TYPE = "job"
MODEL_TYPE = "model"
NOTEBOOK_TYPE = "notebook"
RECIPE_TYPE = "recipe"
TOKEN_TYPE = "token"
USER_TYPE = "user"
WORKSPACE_TYPE = "workspace"
ROLE_TYPE = "role"
DRIVE_TYPE = "drive"

# types/others
PLUGIN_TYPE = "plugin"
WORKER_TYPE = "worker"

# IDs
DATASET_PREFIX = "ds_"  # dataset source, filters, etc
ENDPOINT_PREFIX = "ep_"  # inference endpoint configuration
JOB_PREFIX = "jb_"  # sync or async job
LOG_PREFIX = "lg_"  # log record
MODEL_PREFIX = "ml_"  # trained machine learning model (not a django model)
NOTEBOOK_PREFIX = "nb_"  # Jupyter notebook
RECIPE_PREFIX = "rx_"  # machine learning recipe (an experiment with modules, code, etc)
TOKEN_PREFIX = "tok_"  # authorization token
USER_PREFIX = "id_"  # an identity profile
WORKSPACE_PREFIX = "ws_"  # workspace with rights and one or more projects and other resources
ROLE_PREFIX = "ro_"  # access management rights via roles and permissions
DRIVE_PREFIX = "dr_"  # storage drive, mount, bucket, etc...

PLUGIN_PREFIX = "pl_"  # plugin instance (not plugin description or metadata)
WORKER_PREFIX = "wk_"  # worker process



# status used for logging and jobs
STATUS_CREATED = "created"
STATUS_RUNNING = "running"
STATUS_CANCELED = "canceled"
STATUS_COMPLETED = "completed"
STATUS_FAILED = "failed"
STATUS_ALL = (STATUS_CREATED, STATUS_RUNNING, STATUS_CANCELED, STATUS_COMPLETED, STATUS_FAILED)



import collections
import pandas as pd
import os.path
import multiprocessing
import os
import string
import random

from abc import ABC, abstractmethod

# Design patterns:
# https://github.com/faif/python-patterns

from analitico.mixin import AttributeMixin
from analitico.factory import Factory
from analitico.utilities import time_ms, save_json, read_json, get_runtime_brief
from analitico.schema import apply_schema
from analitico.constants import PLUGIN_PREFIX

##
## IPlugin - base class for all plugins
##


class IPlugin(ABC, AttributeMixin):
    """ Abstract base class for Analitico plugins """

    class Meta:
        """ Plugin metadata is exposed in its inner class """

        name = None

    # Factory that provides runtime services to the plugin (eg: loading assets, etc)
    factory: Factory = None

    @property
    def id(self):
        return self.get_attribute("id")

    @property
    def name(self):
        assert self.Meta.name
        return self.Meta.name

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if "factory" in kwargs:
            self.factory = kwargs["factory"]

    def activate(self, *args, **kwargs):
        """ Called when the plugin is initially activated """
        pass

    @abstractmethod
    def run(self, *args, action=None, **kwargs):
        """ Run will do in the subclass whatever the plugin does, 'action' parameter optional """
        pass

    def deactivate(self, *args, **kwargs):
        """ Called before the plugin is deactivated and finalized """
        pass

    def __str__(self):
        return self.name

    # Utility methods

    def drop_selected_rows(self, df, df_dropped, message=None):
        """ Drops df_dropped rows from dp in place, logs action """
        started_on = time_ms()
        rows_before = len(df.index)
        if rows_before < 1:
            self.warning("Can't drop rows where '%s' because dataframe is empty", message)
            return df
        df.drop(df_dropped.index, inplace=True)
        if message:
            rows_after = len(df.index)
            rows_dropped = rows_before - rows_after
            msg = "Dropped rows where '%s', rows before: %d, after: %d, dropped: %d (%.2f%%) in %d ms"
            self.info(
                msg,
                message,
                rows_before,
                rows_after,
                rows_dropped,
                (100.0 * rows_dropped) / rows_before,
                time_ms(started_on),
            )
        return df

    def drop_na_rows(self, df, column):
        """ Drops rows with null values in given column, logs action """
        started_on = time_ms()
        rows_before = len(df.index)
        if rows_before < 1:
            self.warning("Can't drop null '%s' rows because dataframe is empty", column)
            return df
        df.dropna(subset=[column], inplace=True)
        rows_after = len(df.index)
        rows_dropped = rows_before - rows_after
        msg = "Dropped rows where '%s' is null, rows before: %d, after: %d, dropped: %d (%.2f%%) in %d ms"
        self.info(
            msg,
            column,
            rows_before,
            rows_after,
            rows_dropped,
            (100.0 * rows_dropped) / rows_before,
            time_ms(started_on),
        )
        return df

    # Logging

    @property
    def logger(self):
        """ Logger that can be used by the plugin to communicate errors, etc with host """
        return self.factory.logger

    def info(self, msg, *args, **kwargs):
        self.factory.info(msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        self.factory.warning(msg, *args, **kwargs)

    def error(self, msg, *args, exception=None, **kwargs):
        self.factory.error(msg, *args, plugin=self, exception=exception, **kwargs)

    def exception(self, msg, *args, exception=None, **kwargs):
        self.factory.exception(msg, *args, plugin=self, exception=exception, **kwargs)

    def __str__(self):
        id = self.get_attribute("id")
        return self.Meta.name + ":" + id if id else self.Meta.name


##
## IDataframeSourcePlugin - base class for plugins that create dataframes
##


class IDataframeSourcePlugin(IPlugin):
    """ A plugin that creates a pandas dataframe from a source (eg: csv file, sql query, etc) """

    class Meta(IPlugin.Meta):
        inputs = None
        outputs = [{"name": "dataframe", "type": "pandas.DataFrame"}]

    @abstractmethod
    def run(self, *args, action=None, **kwargs):
        """ Run creates a dataset from the source and returns it """
        pass


##
## IDataframePlugin - base class for plugins that manipulate pandas dataframes
##


class IDataframePlugin(IPlugin):
    """
    A plugin that takes a pandas dataframe as input,
    manipulates it and returns a pandas dataframe
    """

    class Meta(IPlugin.Meta):
        inputs = [{"name": "dataframe", "type": "pandas.DataFrame"}]
        outputs = [{"name": "dataframe", "type": "pandas.DataFrame"}]

    def run(self, *args, action=None, **kwargs) -> pd.DataFrame:
        assert isinstance(args[0], pd.DataFrame)
        return args[0]


##
## IAlgorithmPlugin - base class for machine learning algorithms that produce trained models
##

ALGORITHM_TYPE_REGRESSION = "ml/regression"
ALGORITHM_TYPE_BINARY_CLASSICATION = "ml/binary-classification"
ALGORITHM_TYPE_MULTICLASS_CLASSIFICATION = "ml/multiclass-classification"
ALGORITHM_TYPE_ANOMALY_DETECTION = "ml/anomaly-detection"
ALGORITHM_TYPE_CLUSTERING = "ml/clustering"


class IAlgorithmPlugin(IPlugin):
    """ An algorithm used to create machine learning models from training data """

    class Meta(IPlugin.Meta):
        inputs = [{"name": "train", "type": "pandas.DataFrame"}, {"name": "test", "type": "pandas.DataFrame|none"}]
        outputs = [{"name": "model", "type": "dict"}]

    def _run_train(self, *args, **kwargs):
        """ 
        When an algorithm runs it always takes in a dataframe with training data,
        it may optionally have a dataframe of validation data and will return a dictionary
        with information on the trained model plus a number of artifacts.
        """
        assert isinstance(args[0], pd.DataFrame)
        started_on = time_ms()
        results = collections.OrderedDict(
            {
                "type": "analitico/training",
                "plugins": {
                    "training": self.Meta.name,  # plugin used to train model
                    "prediction": self.Meta.name,  # plugin to be used for predictions (usually the same)
                },
                "data": {},  # number of records, etc
                "parameters": {},  # model parameters, hyperparameters
                "scores": {},  # training scores
                "performance": get_runtime_brief(),  # time elapsed, cpu, gpu, memory, disk, etc
            }
        )

        train = args[0]
        test = args[1] if len(args) > 1 else None
        results = self.train(train, test, results, *args, **kwargs)

        # finalize results and save as training.json
        results["performance"]["total_ms"] = time_ms(started_on)
        artifacts_path = self.factory.get_artifacts_directory()
        results_path = os.path.join(artifacts_path, "training.json")
        save_json(results, results_path)
        self.info("saved %s (%d bytes)", results_path, os.path.getsize(results_path))
        return results

    def _run_predict(self, *args, **kwargs):
        """ 
        When an algorithm runs it always takes in a dataframe with training data,
        it may optionally have a dataframe of validation data and will return a dictionary
        with information on the trained model plus a number of artifacts.
        """
        # assert isinstance(args[0], pandas.DataFrame) # custom models may take json as input
        data = args[0]

        artifacts_path = self.factory.get_artifacts_directory()
        training = read_json(os.path.join(artifacts_path, "training.json"))
        assert training

        started_on = time_ms()
        results = collections.OrderedDict(
            {
                "type": "analitico/prediction",
                # "endpoint_id": None,
                # "model_id": None,
                # "job_id": None,
                # "records": None,  # processed (augmented) data will be added by IAlgorithm
                # "predictions": None,  # predictions
                # "probabilities": None,
                "performance": get_runtime_brief(),  # time elapsed, cpu, gpu, memory, disk, etc
            }
        )

        # force schema like in training data
        if isinstance(data, pd.DataFrame):
            schema = training["data"]["schema"]
            data = apply_schema(data, schema)

        # load model, calculate predictions
        results = self.predict(data, training, results, *args, **kwargs)
        results["performance"]["total_ms"] = time_ms(started_on)

        results_path = os.path.join(artifacts_path, "results.json")
        save_json(results, results_path)

        return results

    def run(self, *args, action=None, **kwargs):
        """ Algorithm can run to train a model or to predict from a trained model """
        if action.endswith("train"):
            return self._run_train(*args, **kwargs)
        if action.endswith("predict"):
            return self._run_predict(*args, **kwargs)
        self.error("unknown action: %s", action)
        raise PluginError("IAlgorithmPlugin - action should be /train or /predict")

    @abstractmethod
    def train(self, train, test, results, *args, **kwargs):
        """ Train with algorithm and given data to produce a trained model """
        pass

    @abstractmethod
    def predict(self, data, training, results, *args, **kwargs):
        """ Return predictions from trained model """
        pass


##
## IGroupPlugin
##


class IGroupPlugin(IPlugin):
    """ 
    A composite plugin that joins multiple plugins into a functional block,
    for example a processing pipeline made of plugins or a graph workflow. 
    
    *References:
    https://en.wikipedia.org/wiki/Composite_pattern
    https://infinitescript.com/2014/10/the-23-gang-of-three-design-patterns/
    """

    plugins = []

    def __init__(self, *args, plugins=[], **kwargs):
        """ Initialize group and create all this plugin's children """
        super().__init__(*args, **kwargs)
        self.plugins = []
        for plugin in plugins:
            if isinstance(plugin, dict):
                plugin = self.factory.get_plugin(**plugin)
            plugin.factory = self.factory  # use same factory as parent plugin
            self.plugins.append(plugin)


##
## PluginError
##


class PluginError(Exception):
    """ Exception generated by a plugin; carries plugin info, inner exception """

    # Plugin error message
    message: str = None

    # Plugin that generated this error (may not be defined)
    plugin: IPlugin = None

    def __init__(self, message, plugin: IPlugin = None, exception: Exception = None):
        super().__init__(message, exception)
        self.message = message
        if plugin:
            self.plugin = plugin
            plugin.logger.error(message)

    def __str__(self):
        if self.plugin:
            return self.plugin.Meta.name + ": " + self.message
        return self.message


##
## @plugin registration helper
##


def plugin(cls):
    """ Use this @plugin decorator on IPlugin classes to register them automatically """
    Factory.register_plugin(cls)
    return cls


def generate_plugin_id():
    """ Generates a random id that is suitable for a plugin instance """
    return PLUGIN_PREFIX + "".join(random.choice(string.hexdigits) for i in range(8))


def apply_plugin_id(plugin_conf: dict):
    """ Checks if the plugin configuration has IDs and if not generates them for this plugin and its children. True if applied, False unchanged. """
    applied = False
    if plugin_conf:
        if "id" not in plugin_conf:
            plugin_conf["id"] = generate_plugin_id()
            applied = True
        if "plugins" in plugin_conf:
            for child_conf in plugin_conf["plugins"]:
                if apply_plugin_id(child_conf):
                    applied = True
    return applied



import unittest
import tempfile
import numpy as np
import os

from analitico.utilities import get_dict_dot, save_json, read_json, read_text, save_text, copy_directory

TST_DICT = {
    "parent_1": {
        "child_1": {"grandchild_1_1": 10, "grandchild_1_2": 20},
        "child_2": {"grandchild_2_1": "30", "grandchild_2_2": "40"},
        "child_3": 24,
    },
    "parent_2": 42,
}


class UtilitiesTests(unittest.TestCase):
    def test_get_dict_dot(self):
        self.assertEqual(get_dict_dot(TST_DICT, "parent_1.child_1.grandchild_1_1"), 10)
        self.assertEqual(get_dict_dot(TST_DICT, "parent_1.child_1.grandchild_1_2"), 20)
        self.assertEqual(get_dict_dot(TST_DICT, "parent_1.child_3"), 24)

    def test_get_dict_dot_missing_value(self):
        self.assertEqual(get_dict_dot(TST_DICT, "parent_4"), None)
        self.assertEqual(get_dict_dot(TST_DICT, "parent_4", 38), 38)
        self.assertEqual(get_dict_dot(TST_DICT, "parent_4", "Mickey"), "Mickey")
        self.assertEqual(get_dict_dot(TST_DICT, "parent_2.unknown"), None)
        self.assertEqual(get_dict_dot(TST_DICT, "parent_2.unknown", 38), 38)
        self.assertEqual(get_dict_dot(TST_DICT, "parent_2.unknown", "Mickey"), "Mickey")
        self.assertEqual(get_dict_dot(TST_DICT, "parent_1.child_3.unknown"), None)
        self.assertEqual(get_dict_dot(TST_DICT, "parent_1.child_3.unknown", 38), 38)
        self.assertEqual(get_dict_dot(TST_DICT, "parent_1.child_3.unknown", "Mickey"), "Mickey")
        self.assertEqual(get_dict_dot(TST_DICT, "parent_2"), 42)

    def test_get_dict_dot_missing_key(self):
        self.assertEqual(get_dict_dot(TST_DICT, None), None)
        self.assertEqual(get_dict_dot(TST_DICT, None, 24), 24)

    def test_save_json(self):
        values1 = {"key1": "value1", "key2": 5}

        with tempfile.NamedTemporaryFile(mode="w+", suffix=".json") as f:
            save_json(values1, f.name)
            values2 = read_json(f.name)

            self.assertEqual(values1["key1"], values2["key1"])
            self.assertEqual(values1["key2"], values2["key2"])

    def test_save_json_with_nan(self):
        """ Test saving a numpy.nan value (should get converted to null) """
        # NaN should be encoded as null, non "NaN" according to ECMA-262
        # https://simplejson.readthedocs.io/en/latest/ (find: ignore_nan)

        values1 = {"nanKey": np.nan}

        with tempfile.NamedTemporaryFile(mode="w+", suffix=".json") as f:
            save_json(values1, f.name, indent=0)

            json2 = read_text(f.name)
            self.assertEqual(json2, '{\n"nanKey": null\n}')

            values2 = read_json(f.name)
            self.assertEqual(values2["nanKey"], None)

    def test_copy_directory(self):
        source = tempfile.TemporaryDirectory()
        destination = tempfile.TemporaryDirectory()
        symlinked_dir = tempfile.TemporaryDirectory()
        symlinked_dir_basename = os.path.basename(symlinked_dir.name)

        f1 = "f1.txt"
        save_text("micky", os.path.join(source.name, f1))
        f2 = "f2.txt"
        save_text("mouse", os.path.join(source.name, f2))
        # duplicate f2 into the destination folder
        save_text("batman", os.path.join(destination.name, f2))
        f3 = "f3.txt"
        save_text("superman",  os.path.join(symlinked_dir.name, f3))
        # create the symlink to the folder
        os.symlink(symlinked_dir.name, os.path.join(source.name, symlinked_dir_basename))

        copy_directory(source.name, destination.name)
        
        # all files are copied
        self.assertTrue(os.path.exists(os.path.join(destination.name, f1)))
        self.assertTrue(os.path.exists(os.path.join(destination.name, f2)))
        self.assertTrue(os.path.exists(os.path.join(destination.name, symlinked_dir_basename, f3)))

        # existing f2 file is ovewritten
        self.assertEqual("mouse", read_text(os.path.join(destination.name, f2)))

        # symbolic link is followed
        copied_symlinked_folder = os.path.join(destination.name, symlinked_dir_basename)
        self.assertFalse(os.path.islink(copied_symlinked_folder))
        self.assertTrue(os.path.exists(os.path.join(copied_symlinked_folder, f3)))

        # missing destination directory is created
        with tempfile.TemporaryDirectory() as temp:
            destination_missing = os.path.join(temp, "subfolder")
            copy_directory(source.name, destination_missing)
            self.assertTrue(os.path.exists(destination_missing))




import unittest
import os
import os.path
import pytest
import pandas as pd

from analitico.plugin import PluginError, PLUGIN_TYPE
from analitico.plugin import CsvDataframeSourcePlugin, CSV_DATAFRAME_SOURCE_PLUGIN
from analitico.plugin import CODE_DATAFRAME_PLUGIN
from analitico.plugin import PipelinePlugin, PIPELINE_PLUGIN

from .test_mixin import TestMixin

# pylint: disable=no-member

ASSETS_PATH = os.path.dirname(os.path.realpath(__file__)) + "/assets"


@pytest.mark.django_db
class PluginTests(unittest.TestCase, TestMixin):
    """ Unit testing of Plugin functionalities """

    def test_plugin_basics_settings(self):
        """ Test plugin settings """
        try:
            plugin = CsvDataframeSourcePlugin(factory=self.factory, param1="value1", param2="value2")

            self.assertEqual(plugin.param1, "value1")
            self.assertEqual(plugin.param2, "value2")

            self.assertEqual(plugin.get_attribute("param1"), "value1")
            self.assertEqual(plugin.get_attribute("param2"), "value2")
        except Exception as exc:
            raise exc

    def test_plugin_factory(self):
        try:
            plugin = self.factory.get_plugin(CSV_DATAFRAME_SOURCE_PLUGIN, param1="value1", param2="value2")

            self.assertEqual(plugin.param1, "value1")
            self.assertEqual(plugin.param2, "value2")
            self.assertEqual(plugin.get_attribute("param1"), "value1")
            self.assertEqual(plugin.get_attribute("param2"), "value2")
        except Exception as exc:
            raise exc

    def test_plugin_csv_read(self):
        """ Test using csv plugin to read basic csv file """
        csv_url = self.get_asset_path("ds_test_1.csv")
        csv_plugin = self.get_csv_plugin(source={"url": csv_url})
        self.assertTrue(isinstance(csv_plugin, CsvDataframeSourcePlugin))
        self.assertEqual(csv_plugin.get_attribute("source.url"), csv_url)

        df = csv_plugin.run()
        self.assertTrue(isinstance(df, pd.DataFrame))
        self.assertIsNotNone(df)
        self.assertEqual(len(df), 3)

    def test_plugin_code_dataframe_basic(self):
        """ Test using csv plugin to applies basic code to a dataframe """
        csv_url = self.get_asset_path("ds_test_1.csv")
        csv_plugin = self.get_csv_plugin(source={"url": csv_url})

        df = csv_plugin.run()
        self.assertEqual(df.loc[0, "First"], 10)

        # configure plugin to add 2 to all values in the first column of the dataframe
        code = "df['First'] = df['First'] + 2"
        plugin = self.factory.get_plugin(CODE_DATAFRAME_PLUGIN, code=code)

        # dataframe passed as POSITIONAL parameter
        df = plugin.run(df, action="dataset/process")
        self.assertEqual(df.loc[0, "First"], 12)

        # dataframe passed as POSITIONAL parameter
        df = plugin.run(df, action="dataset/process")
        self.assertEqual(df.loc[0, "First"], 14)

    def test_plugin_code_dataframe_bug(self):
        """ Test using csv plugin to applies code with a bug to a dataframe """
        csv_url = self.get_asset_path("ds_test_1.csv")
        csv_plugin = self.factory.get_plugin(CSV_DATAFRAME_SOURCE_PLUGIN, source={"url": csv_url})

        df = csv_plugin.run()
        self.assertEqual(df.loc[0, "First"], 10)

        # refers to df2 which DOES NOT exist
        code = "df['First'] = INTENTIONALLY_UNDEFINED_df2['First'] + 2"
        plugin = self.factory.get_plugin(CODE_DATAFRAME_PLUGIN, code=code)

        with self.assertRaises(PluginError):
            df = plugin.run(df, actions="dataset/process")

    def test_plugin_pipeline(self):
        """ Test grouping plugins into a multi step pipeline to retrieve and process a dataframe """
        pipeline_settings = {
            "type": PLUGIN_TYPE,
            "name": PIPELINE_PLUGIN,
            "plugins": [
                {
                    "type": PLUGIN_TYPE,
                    "name": CSV_DATAFRAME_SOURCE_PLUGIN,
                    "source": {"url": self.get_asset_path("ds_test_1.csv")},
                },
                {"type": PLUGIN_TYPE, "name": CODE_DATAFRAME_PLUGIN, "code": "df['First'] = df['First'] + 2"},
                {"type": PLUGIN_TYPE, "name": CODE_DATAFRAME_PLUGIN, "code": "df['First'] = df['First'] + 4"},
                {"type": PLUGIN_TYPE, "name": CODE_DATAFRAME_PLUGIN, "code": "df['First'] = df['First'] + 1"},
            ],
        }

        pipeline_plugin = self.factory.get_plugin(**pipeline_settings)
        self.assertTrue(isinstance(pipeline_plugin, PipelinePlugin))

        # call plugin chain; pass same random parameters just to see that they don't mess up things
        pipeline_df = pipeline_plugin.run("par1", "par2", mickey="mouse", minni="pluto")
        self.assertIsNotNone(pipeline_df)
        self.assertTrue(isinstance(pipeline_df, pd.DataFrame))

        # plugin chain increased first column by 2 + 4 + 1
        self.assertEqual(pipeline_df.loc[0, "First"], 17)
        self.assertEqual(pipeline_df.loc[1, "First"], 27)
        # second column untouched
        self.assertEqual(pipeline_df.loc[0, "Second"], 11)
        self.assertEqual(pipeline_df.loc[1, "Second"], 21)

        # call plugin chain again with some random positional
        # and named parameters just to see that they don't mess up things
        # parameters should be passed down the chain of plugins and ignored
        pipeline_df2 = pipeline_plugin.run("par1", "par2", mickey="mouse", minni="pluto")
        self.assertIsNotNone(pipeline_df2)
        self.assertTrue(isinstance(pipeline_df2, pd.DataFrame))

        # plugin chain increased first column by 2 + 4 + 1
        self.assertEqual(pipeline_df2.loc[0, "First"], 17)
        self.assertEqual(pipeline_df2.loc[1, "First"], 27)
        # second column untouched
        self.assertEqual(pipeline_df2.loc[0, "Second"], 11)
        self.assertEqual(pipeline_df2.loc[1, "Second"], 21)



import os
import logging

# default logger used by libraries, etc
logger = logging.getLogger("analitico")

from .constants import *
from .exceptions import *

import analitico.factory
import analitico.utilities
import analitico.mixin
import analitico.plugin
import analitico.dataset
import analitico.status
import analitico.logging

import analitico.sdk

# classes used to represent items in the service
from analitico.models import Item, Dataset, Recipe, Notebook


# from analitico.factory import Item, Dataset, Recipe, Notebook, AnaliticoException

# import utility methods to main namespace
from analitico.metadata import set_metric, set_model_metrics

# DEPRECATED
def authorize(token=None, endpoint=ANALITICO_API_ENDPOINT) -> analitico.factory.Factory:
    """ 
    Returns an API factory which can create datasets, models, run notebooks, plugins, etc.
    You can pass an API token as a parameter or you can set the ANALITICO_API_ENDPOINT environment
    variable with a token that should be used to authorize API calls. By default calls will be
    made to the production environment but you can specify staging or any other endpoint.
    """
    if not token:
        token = os.environ.get("ANALITICO_API_TOKEN", None)
        if not token:
            logger.warning(
                "authorize - you should pass an API token or set the environment variable ANALITICO_API_ENDPOINT"
            )
    return analitico.factory.Factory(token=token, endpoint=endpoint)


#
def authorize_sdk(token=None, endpoint=ANALITICO_API_ENDPOINT, workspace_id: str = None) -> analitico.factory.Factory:
    """ 
    Returns an API factory which can create datasets, models, run notebooks, plugins, etc.
    You can pass an API token as a parameter or you can set the ANALITICO_API_ENDPOINT environment
    variable with a token that should be used to authorize API calls. By default calls will be
    made to the production environment but you can specify staging or any other endpoint.
    """
    if not token:
        token = os.environ.get("ANALITICO_API_TOKEN", None)
        if not token:
            logger.warning(
                "authorize - you should pass an API token or set the environment variable ANALITICO_API_ENDPOINT"
            )
    return analitico.sdk.AnaliticoSDK(token=token, endpoint=endpoint, workspace_id=workspace_id)



# plugin base classes
from .interfaces import *

# plugins to generate dataframes from sources
from .csvdataframesourceplugin import CsvDataframeSourcePlugin
from .datasetsourceplugin import DatasetSourcePlugin

# plugins to tranform dataframes
from .transforms import CodeDataframePlugin
from .augmentdatesplugin import AugmentDatesPlugin
from .fusiondataframeplugin import FusionDataframePlugin
from .transformdataframeplugin import TransformDataframePlugin

# machine learning algorithms
from .catboostplugin import CatBoostPlugin
from .catboostplugin import CatBoostRegressorPlugin
from .catboostplugin import CatBoostClassifierPlugin

# plugin workflows
from .pipelineplugin import PipelinePlugin
from .dataframepipelineplugin import DataframePipelinePlugin
from .recipepipelineplugin import RecipePipelinePlugin
from .endpointpipelineplugin import EndpointPipelinePlugin

CSV_DATAFRAME_SOURCE_PLUGIN = CsvDataframeSourcePlugin.Meta.name
DATASET_SOURCE_PLUGIN = DatasetSourcePlugin.Meta.name
CODE_DATAFRAME_PLUGIN = CodeDataframePlugin.Meta.name
AUGMENT_DATES_PLUGIN = AugmentDatesPlugin.Meta.name
FUSION_DATAFRAME_PLUGIN = FusionDataframePlugin.Meta.name
TRANSFORM_DATAFRAME_PLUGIN = TransformDataframePlugin.Meta.name
CATBOOST_PLUGIN = CatBoostPlugin.Meta.name
CATBOOST_REGRESSOR_PLUGIN = CatBoostRegressorPlugin.Meta.name
CATBOOST_CLASSIFIER_PLUGIN = CatBoostClassifierPlugin.Meta.name
PIPELINE_PLUGIN = PipelinePlugin.Meta.name
DATAFRAME_PIPELINE_PLUGIN = DataframePipelinePlugin.Meta.name
RECIPE_PIPELINE_PLUGIN = RecipePipelinePlugin.Meta.name
ENDPOINT_PIPELINE_PLUGIN = EndpointPipelinePlugin.Meta.name

# analitico type for plugins
PLUGIN_TYPE = "analitico/plugin"

# NOQA: F401 prospector complains that these imports
# are unused but they are here to define the module



import unittest
import os.path
import pandas as pd
import random
import string
import pytest

from analitico import AnaliticoException
from analitico.factory import Factory
from analitico.schema import generate_schema

from .test_mixin import TestMixin

# pylint: disable=no-member


@pytest.mark.django_db
class ExceptionTests(unittest.TestCase, TestMixin):
    """ Unit testing of AnaliticoException """

    def test_exception_basics(self):
        try:
            raise AnaliticoException("A problem", item_id="it_001", extra={"recipe_id": "rx_002"})
        except AnaliticoException as e:
            self.assertEqual(e.message, "A problem")
            self.assertEqual(e.extra["item_id"], "it_001")
            self.assertEqual(e.extra["recipe_id"], "rx_002")

    def test_exception_formatting(self):
        try:
            raise AnaliticoException("A problem %d", 1, item_id="it_001", extra={"recipe_id": "rx_002"})
        except AnaliticoException as e:
            self.assertEqual(e.message, "A problem 1")
            self.assertEqual(e.extra["item_id"], "it_001")
            self.assertEqual(e.extra["recipe_id"], "rx_002")

    def test_exception_formatting_multiple(self):
        try:
            raise AnaliticoException("A problem %d with %s", 1, "this", item_id="it_001", extra={"recipe_id": "rx_002"})
        except AnaliticoException as e:
            self.assertEqual(e.message, "A problem 1 with this")

    def test_exception_with_inner(self):
        try:
            try:
                raise AnaliticoException("Inner exception %d", 1, item_id="it_001", extra={"recipe_id": "rx_002"})
            except AnaliticoException as e1:
                self.assertEqual(e1.message, "Inner exception 1")
                self.assertEqual(e1.extra["item_id"], "it_001")
                self.assertEqual(e1.extra["recipe_id"], "rx_002")
                raise AnaliticoException("Outer exception %d", 2, test_id="tx001") from e1
        except AnaliticoException as e2:
            self.assertEqual(e2.message, "Outer exception 2")
            print(e2)



import logging
import logging.config
import time
import json

from datetime import datetime

# Adapted from marselester/json-log-formatter:
# https://github.com/marselester/json-log-formatter

BUILTIN_ATTRS = {
    "args",
    "asctime",
    "created",
    "exc_info",
    "exc_text",
    "filename",
    "funcName",
    "levelname",
    "levelno",
    "lineno",
    "module",
    "msecs",
    "message",
    "msg",
    "name",
    "pathname",
    "process",
    "processName",
    "relativeCreated",
    "stack_info",
    "thread",
    "threadName",
}


class FluentdFormatter(logging.Formatter):
    """
    A log formatter used to collect logs on a kubernets cluster in a way that
    can be easily collected by fluentd deamons and the dispatched to elastich
    search so it can be visualized in analitico's dashboard with reach metadata.

    Usage example::
        import logging
        import json_log_formatter
        json_handler = logging.FileHandler(filename='/var/log/my-log.json')
        json_handler.setFormatter(json_log_formatter.FluentdFormatter())
        logger = logging.getLogger('my_json')
        logger.addHandler(json_handler)
        logger.info('Sign up', extra={'referral_code': '52d6ce'})

    The log file will contain the following log record (inline)::
        {
            "level": "warning", 
            "ts": "2019-07-03T10:27:59.430590", 
            "logger": "root", 
            "caller": "<module>:33", 
            "msg": "Sign up",
            "referral_code": "52d6ce"
        }
    """

    json_lib = json

    def format(self, record):
        message = record.getMessage()
        extra = self.extra_from_record(record)
        json_record = self.json_record(message, extra, record)
        mutated_record = self.mutate_json_record(json_record)
        # Backwards compatibility: Functions that overwrite this but don't
        # return a new value will return None because they modified the
        # argument passed in.
        if mutated_record is None:
            mutated_record = json_record
        return self.to_json(mutated_record)

    def to_json(self, record):
        """ 
        Converts record dict to a JSON string. 
        Override this method to change the way dict is converted to JSON. 
        """
        return self.json_lib.dumps(record)

    def extra_from_record(self, record):
        """
        Returns `extra` dict you passed to logger.
        The `extra` keyword argument is used to populate the `__dict__` of the `LogRecord`.
        """
        return {
            attr_name: record.__dict__[attr_name] for attr_name in record.__dict__ if attr_name not in BUILTIN_ATTRS
        }

    def json_record(self, message, extra, record):
        """
        Prepares a JSON payload which will be logged.
        Override this method to change JSON log format.
        :param message: Log message, e.g., `logger.info(msg='Sign up')`.
        :param extra: Dictionary that was passed as `extra` param
            `logger.info('Sign up', extra={'referral_code': '52d6ce'})`.
        :param record: `LogRecord` we got from `JSONFormatter.format()`.
        :return: Dictionary which will be passed to JSON lib.
        """
        extra["level"] = record.levelname.lower()
        if "ts" not in extra:
            extra["ts"] = datetime.utcnow()
        extra["logger"] = record.name
        if record.funcName and record.lineno:
            extra["caller"] = record.funcName + ":" + str(record.lineno)
        extra["msg"] = message
        if record.exc_info:
            extra["exc_info"] = self.formatException(record.exc_info)
        return extra

    def mutate_json_record(self, json_record):
        """
        Override it to convert fields of `json_record` to needed types.
        Default implementation converts `datetime` to string in ISO8601 format.
        """
        for attr_name in json_record:
            attr = json_record[attr_name]
            if isinstance(attr, datetime):
                json_record[attr_name] = attr.isoformat()
        return json_record



import analitico
from analitico import AnaliticoException

from .item import Item
from .workspace import Workspace
from .dataset import Dataset
from .recipe import Recipe
from .notebook import Notebook


def models_factory(sdk, item_data: dict) -> Item:
    """
    Create an analitico model from the json response returned by the service.
    
    Arguments:
        item_data {dict} -- Json returned by service. Should contain id, type, attributes.
    
    Returns:
        Item -- An item, eg: Dataset, Recipe, Notebook or generic Item.
    """
    if not "type" in item_data:
        raise AnaliticoException(
            "An item should have a type field, eg: analitico/dataset, analitico/recipe, analitico/notebook, etc."
        )

    if item_data["type"] == "analitico/" + analitico.WORKSPACE_TYPE:
        return Workspace(sdk, item_data)
    if item_data["type"] == "analitico/" + analitico.DATASET_TYPE:
        return Dataset(sdk, item_data)
    if item_data["type"] == "analitico/" + analitico.RECIPE_TYPE:
        return Recipe(sdk, item_data)
    if item_data["type"] == "analitico/" + analitico.NOTEBOOK_TYPE:
        return Notebook(sdk, item_data)

    # generic untyped item
    return Item(sdk, item_data)



""" Utility methods to convert between pandas and analitico's schemas """

import numpy as np
import pandas as pd

##
## Schema
##

# pandas types for analitico's types
PD_TYPE_INTEGER = "int64"
PD_TYPE_FLOAT = "float64"
PD_TYPE_STRING = "str"
PD_TYPE_BOOLEAN = "bool"
PD_TYPE_DATETIME = "datetime64"
PD_TYPE_TIMESPAN = "timedelta64"
PD_TYPE_CATEGORY = "category"

ANALITICO_TYPE_INTEGER = "integer"
ANALITICO_TYPE_FLOAT = "float"
ANALITICO_TYPE_STRING = "string"
ANALITICO_TYPE_BOOLEAN = "boolean"
ANALITICO_TYPE_DATETIME = "datetime"
ANALITICO_TYPE_TIMESPAN = "timespan"
ANALITICO_TYPE_CATEGORY = "category"

# these values are replaced with pd.NaN
NA_VALUES = [
    "None",
    "",
    "#N/A",
    "#N/A",
    "N/A",
    "#NA",
    "-1.#IND",
    "-1.#QNAN",
    "-NaN",
    "-nan",
    "1.#IND",
    "1.#QNAN",
    "N/A",
    "NA",
    "NULL",
    "NaN",
    "n/a",
    "nan",
    "null",
]

# these values are replaced with pd.NaT
NA_DATES = NA_VALUES + [0, "0"]


def analitico_to_pandas_type(data_type: str):
    """ Converts an analitico data type to the equivalent dtype string for pandas dataframes """
    try:
        ANALITICO_TO_PANDAS_TYPES = {
            ANALITICO_TYPE_STRING: PD_TYPE_STRING,
            ANALITICO_TYPE_INTEGER: PD_TYPE_INTEGER,
            ANALITICO_TYPE_FLOAT: PD_TYPE_FLOAT,
            ANALITICO_TYPE_BOOLEAN: PD_TYPE_BOOLEAN,
            ANALITICO_TYPE_DATETIME: PD_TYPE_DATETIME,
            ANALITICO_TYPE_TIMESPAN: PD_TYPE_TIMESPAN,
            ANALITICO_TYPE_CATEGORY: PD_TYPE_CATEGORY,
        }
        return ANALITICO_TO_PANDAS_TYPES[data_type]
    except KeyError as exc:
        raise KeyError("analitico_to_pandas_type - unknown type: " + data_type, exc)


def pandas_to_analitico_type(data_type):
    """ Return the analitico schema data type of a pandas dtype """
    if data_type == "int" or data_type == "int8":
        return ANALITICO_TYPE_INTEGER
    if data_type == "float":
        return ANALITICO_TYPE_FLOAT
    if data_type == "bool":
        return ANALITICO_TYPE_BOOLEAN
    if data_type.name == "category":
        return ANALITICO_TYPE_CATEGORY  # dtype alone doesn't ==
    if data_type == "object":
        return ANALITICO_TYPE_STRING
    if data_type == "datetime64[ns]":
        return ANALITICO_TYPE_DATETIME
    if data_type == "timedelta64[ns]":
        return ANALITICO_TYPE_TIMESPAN
    raise KeyError("_pandas_to_analitico_type - unknown data_type: " + str(data_type))


def get_column_type(df, column):
    """ Returns analitico's column type for the column with given name """
    return pandas_to_analitico_type(df[column].dtype)


def generate_schema(df: pd.DataFrame) -> dict:
    """ Generates an analitico schema from a pandas dataframe """
    columns = []

    columns_names = df.columns.tolist()
    for name in columns_names:
        ctype = pandas_to_analitico_type(df[name].dtype)
        column = {"name": name, "type": ctype}
        if df.index.name == name:
            column["index"] = True
        columns.append(column)
    return {"columns": columns}


def apply_column(df: pd.DataFrame, column):
    """ Apply given type to the column (parameters are type, name, etc from schema column) """
    assert isinstance(df, pd.DataFrame)
    assert "name" in column, "apply_column - should always be passed a column name"
    column_name = column["name"]

    # we are being requested to apply type to the column
    if "type" in column:
        try:
            column_type = column["type"]
            missing = column_name not in df.columns
            if column_type == "string":
                if missing:
                    df[column_name] = None
                df[column_name] = df[column_name].astype(str)
            elif column_type == "float":
                if missing:
                    df[column_name] = np.nan
                df[column_name].fillna(0)
                df[column_name] = df[column_name].astype(float)
            elif column_type == "boolean":
                if missing:
                    df[column_name] = False
                # missing values are converted to False
                df[column_name] = df[column_name].fillna(False).astype(bool)
            elif column_type == "integer":
                if missing:
                    df[column_name] = 0
                # missing values converted to 0
                df[column_name] = df[column_name].fillna(0).astype(int)
            elif column_type == "datetime":
                if missing:
                    df[column_name] = None
                else:
                    # strings like no
                    for not_a_date in NA_DATES:
                        df[column_name].replace(not_a_date, np.nan, inplace=True)
                df[column_name] = df[column_name].astype("datetime64[ns]")
            elif column_type == "timespan":
                if missing:
                    df[column_name] = None
                df[column_name] = pd.to_timedelta(df[column_name])
            elif column_type == "category":
                if missing:
                    df[column_name] = None
                df[column_name] = df[column_name].astype("category")
            else:
                raise Exception("apply_column - unknown type: " + column_type)
        except Exception as exc:
            raise Exception(
                "apply_column - exception while applying type '" + column_type + "' to column '" + column_name + "'"
            ) from exc

    if "rename" in column:
        df.rename(index=str, columns={column_name: column["rename"]}, inplace=True)
        column_name = column["rename"]

    # make requested column index
    index = column.get("index", False)
    if index:
        # we use this column as the index but do not remove it from
        # the columns otherwise we won't be able to rename it, etc
        df.set_index(column_name, drop=False, inplace=True)

    assert column_name in df.columns
    return df[column_name]


def apply_schema(df: pd.DataFrame, schema):
    """ 
    Applies the given schema to the dataframe. The method will scan columns
    in the schema and apply their type to columns in the dataframe. It will
    then sort, filter and rename columns according to schema.
    """
    assert isinstance(df, pd.DataFrame), "apply_schema should be passed a pd.DataFrame, received: " + str(df)
    assert isinstance(schema, dict), "apply_schema should be passed a schema dictionary"

    # when a schema contains the 'columns' array, it means that we should
    # apply the given columns transformations and end up with a dataframe
    # containing only the given columns, ordered as specified. if the dataframe
    # has any additional columns not listed in 'columns' they will be dropped.
    if "columns" in schema:
        # select columns and apply types to columns
        names = []
        for column in schema["columns"]:
            ds = apply_column(df, column)
            names.append(ds.name)
        # reorder and remove extra columns
        return df[names]

    # when a schema contains the 'apply' array it means that we should
    # apply the given column transformations to the dataframe. the array
    # may or just one or a few columns, meaning all other columns are left
    # as is. the order of items in 'apply' does not need to match the order
    # in the dataframe
    if "apply" in schema:
        for column in schema["apply"]:
            apply_column(df, column)

    # if schema has a "drop" array then the columns
    if "drop" in schema:
        for column in schema["drop"]:
            column_name = column.get("name")
            if column_name and column_name in df.columns:
                df.drop(columns=[column_name], inplace=True)

    return df



import pandas

from .interfaces import IDataframePlugin, PluginError, plugin

##
## CodeDataframePlugin
##


@plugin
class CodeDataframePlugin(IDataframePlugin):
    """
    A plugin that can apply some generic python code to a dataframe and return it.
    Normally a short bit of code is used to apply expressions using pandas or to filter
    rows and such. The dataframe can be accessed in the code snippet using the variable
    'df' and returned in the same variable. The code snipped is passed to the plugin
    using the setting 'code' containing the code itself. This plugin is not isolating
    the code therefore it should only run internal code or it will expose a security risk.
    Later on we will create a version of this plugin that uses dockers to isolate the code.
    """

    class Meta(IDataframePlugin.Meta):
        name = "analitico.plugin.CodeDataframePlugin"

    def run(self, *args, action=None, **kwargs) -> pandas.DataFrame:
        """ Apply some python code to the dataframe """
        df = super().run(*args, action=action, **kwargs)
        code = self.get_attribute("code", None)
        if code:
            try:
                # TODO plugin should restrict code execution to math, numpy and pandas #17
                # https://www.programiz.com/python-programming/methods/built-in/exec
                exec(code)
            except Exception as exc:
                message = 'Error while executing "{0}": "{1}".'.format(code, exc)
                self.logger.error(message)
                raise PluginError(message, self, exc)
        return df



import unittest
import pytest
import pandas as pd

from analitico.schema import generate_schema, apply_schema

from .test_mixin import TestMixin

# pylint: disable=no-member


@pytest.mark.django_db
class DatasetTests(unittest.TestCase, TestMixin):
    """ Unit testing of Dataset functionality, reading, converting, transforms, saving, etc """

    ## Test creations

    def test_dataset_csv1_basics(self):
        """ Test empty dataset creation """
        try:
            ds = self.read_dataset_asset("ds_test_1.json")
            self.assertEqual(ds.id, "ds_1")

            df = ds.get_dataframe()
            self.assertTrue(isinstance(df, pd.DataFrame))
            self.assertEqual(len(df), 3)
            self.assertEqual(df.columns[0], "First")
            self.assertEqual(df.columns[1], "Second")
            self.assertEqual(df.columns[2], "Third")
            self.assertEqual(df.iloc[0, 0], 10)
            self.assertEqual(df.iloc[1, 1], 21)
            self.assertEqual(df.iloc[2, 2], 32)
        except Exception as exc:
            raise exc

    def test_dataset_csv2_types_default(self):
        """ Test standard data type conversions """
        try:
            ds = self.read_dataset_asset("ds_test_2.json")
            self.assertEqual(ds.id, "ds_2")

            df = ds.get_dataframe()
            self.assertEqual(df.dtypes[0], "int64")
            self.assertEqual(df.dtypes[1], "O")
            self.assertEqual(df.dtypes[2], "float64")
        except Exception as exc:
            raise exc

    def test_dataset_csv3_types_cast_float(self):
        """ Test forcing integers to be floating point instead """
        try:
            df = self.read_dataframe_asset("ds_test_3_cast_float.json")
            # would normally be int, but was forced to float
            self.assertEqual(df.dtypes[0], "float64")
            self.assertEqual(df.dtypes[1], "O")
            self.assertEqual(df.dtypes[2], "float64")
        except Exception as exc:
            raise exc

    def test_dataset_csv3_types_cast_string(self):
        """ Test forcing float column to string """
        try:
            df = self.read_dataframe_asset("ds_test_3_cast_string.json")
            self.assertEqual(df.dtypes[0], "int64")
            self.assertEqual(df.dtypes[1], "O")
            # third column would be float, but is cast to string
            self.assertEqual(df.dtypes[2], "O")
            self.assertEqual(df.iloc[2, 2], "32.50")
        except Exception as exc:
            raise exc

    def test_dataset_csv4_applyschema_rename(self):
        """ Test reading a table then renaming a column """
        try:
            df = self.read_dataframe_asset("ds_test_4.json")
            schema = generate_schema(df)

            columns = schema["columns"]
            self.assertEqual(len(columns), 3)
            self.assertEqual(df.columns[1], "Second")

            schema["columns"][1]["rename"] = "Secondo"
            df = apply_schema(df, schema)

            columns = df.columns
            self.assertEqual(df.columns[1], "Secondo")
        except Exception as exc:
            raise exc

    def test_dataset_csv4_applyschema_index(self):
        """ Test reading a table then making a column its index """
        try:
            df = self.read_dataframe_asset("ds_test_4.json")
            schema = generate_schema(df)

            columns = schema["columns"]
            self.assertEqual(len(columns), 3)
            self.assertEqual(df.index.name, None)

            schema["columns"][0]["index"] = True
            df = apply_schema(df, schema)

            columns = df.columns
            self.assertEqual(df.index.name, "First")
        except Exception as exc:
            raise exc

    def test_dataset_csv4_applyschema_index_rename(self):
        """ Test reading a table then making a column its index then renaming it """
        try:
            df = self.read_dataframe_asset("ds_test_4.json")
            schema = generate_schema(df)

            columns = schema["columns"]
            self.assertEqual(len(columns), 3)
            self.assertEqual(df.index.name, None)

            schema["columns"][0]["index"] = True
            schema["columns"][0]["rename"] = "Primo"
            df = apply_schema(df, schema)

            columns = df.columns
            self.assertEqual(df.index.name, "Primo")
            self.assertEqual(df.columns[0], "Primo")
        except Exception as exc:
            raise exc

    def test_dataset_csv4_types_datetime_iso8601(self):
        """ Test reading datetime in ISO8601 format """
        try:
            df = self.read_dataframe_asset("ds_test_4.json")
            self.assertEqual(df.dtypes[0], "int64")
            self.assertEqual(df.dtypes[1], "O")

            self.assertTrue(isinstance(df.iloc[0, 2], pd.Timestamp))
            self.assertTrue(isinstance(df.iloc[1, 2], pd.Timestamp))
            self.assertTrue(isinstance(df.iloc[2, 2], pd.Timestamp))
            self.assertTrue(isinstance(df.iloc[3, 2], pd.Timestamp))

            self.assertEqual(df.iloc[0, 2], pd.Timestamp("2019-01-20 00:00:00"))
            self.assertEqual(df.iloc[1, 2], pd.Timestamp("2019-01-20 16:30:15"))
            self.assertEqual(df.iloc[2, 2], pd.Timestamp("2019-02-01 00:00:00"))
            self.assertEqual(df.iloc[3, 2], pd.Timestamp("2019-01-01 00:00:00"))

            # Timezones are state machines from row to row...

            # 2019-09-15T15:53:00
            self.assertEqual(df.iloc[4, 2], pd.Timestamp("2019-09-15 15:53:00"))
            # 2019-09-15T15:53:00+05:00 (changes timezone)
            self.assertEqual(df.iloc[5, 2], pd.Timestamp("2019-09-15 10:53:00"))
            # 2019-09-15T15:53:00 (maintains +5 timezone)
            self.assertEqual(df.iloc[6, 2], pd.Timestamp("2019-09-15 10:53:00"))
            # 2019-09-15T15:53:00+00 (reverts timezone)
            self.assertEqual(df.iloc[7, 2], pd.Timestamp("2019-09-15 15:53:00"))
            # 2019-09-15T15:53:00-01:30 (changes timezone)
            self.assertEqual(df.iloc[8, 2], pd.Timestamp("2019-09-15 17:23:00"))
            # 20080915T155300Z (zulu time)
            self.assertEqual(df.iloc[9, 2], pd.Timestamp("2008-09-15 15:53:00"))

            # Time only uses today's date: 15:53:00.322348
            self.assertEqual(df.iloc[10, 2], pd.Timestamp("15:53:00.322348"))

            # Examples:
            # http://support.sas.com/documentation/cdl/en/lrdict/64316/HTML/default/viewer.htm#a003169814.htm
        except Exception as exc:
            raise exc

    def test_dataset_csv5_category_no_schema(self):
        """ Test reading categorical data without a schema """
        try:
            df = self.read_dataframe_asset("ds_test_5_category_no_schema.json")

            self.assertEqual(len(df.columns), 10)
            self.assertEqual(df.columns[0], "id")
            self.assertEqual(df.columns[1], "name")
            self.assertEqual(df.columns[2], "slug")
            self.assertEqual(df.columns[3], "parent_id")
            self.assertEqual(df.columns[4], "depth")
            self.assertEqual(df.columns[5], "priority")
            self.assertEqual(df.columns[6], "max_weight")
            self.assertEqual(df.columns[7], "frozen")
            self.assertEqual(df.columns[8], "rate")
            self.assertEqual(df.columns[9], "has_ingredients_book")

            # Column types
            self.assertEqual(df.dtypes[0], "int")  # id
            self.assertEqual(df.dtypes[1], "O")  # name
            self.assertEqual(df.dtypes[2], "O")  # slug
            self.assertEqual(df.dtypes[3], "float")  # parent_id
            self.assertEqual(df.dtypes[7], "int")  # frozen

            # Items types
            self.assertEqual(type(df.iloc[0, 1]).__name__, "str")  # name
            self.assertEqual(type(df.iloc[0, 2]).__name__, "str")  # slug
            self.assertEqual(type(df.iloc[0, 3]).__name__, "float64")  # parent_id
        except Exception as exc:
            raise exc

    def test_dataset_csv5_category_with_schema(self):
        """ Test reading categorical data with a schema, check types """
        try:
            df = self.read_dataframe_asset("ds_test_5_category_with_schema.json")

            self.assertEqual(len(df.columns), 10)
            self.assertEqual(df.columns[0], "id")
            self.assertEqual(df.columns[1], "name")
            self.assertEqual(df.columns[2], "slug")
            self.assertEqual(df.columns[3], "parent_id")
            self.assertEqual(df.columns[4], "depth")
            self.assertEqual(df.columns[5], "priority")
            self.assertEqual(df.columns[6], "max_weight")
            self.assertEqual(df.columns[7], "frozen")
            self.assertEqual(df.columns[8], "rate")
            self.assertEqual(df.columns[9], "has_ingredients_book")

            # Column types
            self.assertEqual(df.dtypes[0], "int")  # id
            self.assertEqual(df.dtypes[1], "category")  # name
            self.assertEqual(df.dtypes[2], "category")  # slug
            self.assertEqual(df.dtypes[3], "float")  # parent_id
            self.assertEqual(df.dtypes[7], "bool")  # frozen
        except Exception as exc:
            raise exc

    def test_dataset_csv5_category_check_values(self):
        """ Test reading categorical data, check values """
        try:
            df = self.read_dataframe_asset("ds_test_5_category_with_schema.json")

            # Items types
            self.assertEqual(type(df.iloc[0, 1]).__name__, "str")  # name
            self.assertEqual(type(df.iloc[0, 2]).__name__, "str")  # slug
            self.assertEqual(type(df.iloc[0, 3]).__name__, "float64")  # parent_id
            self.assertEqual(type(df.iloc[0, 7]).__name__, "bool_")  # frozen
        except Exception as exc:
            raise exc

    def test_dataset_csv5_category_no_index(self):
        """ Test reading categorical data, check index column """
        try:
            df1 = self.read_dataframe_asset("ds_test_5_category_with_schema.json")

            # By default the index column is the row number.
            # If the dataset has an index or id row it is just like
            # any other row and is not used to index the pandas dataset
            self.assertFalse(df1.loc[205, "frozen"])
            self.assertEqual(df1.loc[205, "slug"], "sughi-pronti-primi-piatti")
            self.assertEqual(df1.loc[205, "parent_id"], 100150)

            # Apply the correct index column manually
            df2 = df1.set_index("id", drop=False)
            self.assertFalse(df2.loc[205, "frozen"])
            self.assertEqual(df2.loc[205, "slug"], "carne-tacchino")
            self.assertEqual(df2.loc[205, "parent_id"], 100102)
        except Exception as exc:
            raise exc

    def test_dataset_csv5_category_with_index(self):
        """ Test reading categorical data, check explicit index column """
        try:
            df = self.read_dataframe_asset("ds_test_5_category_with_index.json")
            self.assertFalse(df.loc[205, "frozen"])
            self.assertEqual(df.loc[205, "slug"], "carne-tacchino")
            self.assertEqual(df.loc[205, "parent_id"], 100102)
        except Exception as exc:
            raise exc

    def test_dataset_csv6_weird_index_no_attr(self):
        """ Test reading table with 'weird' index column explicitly marked in schema """
        try:
            df = self.read_dataframe_asset("ds_test_6_weird_index_no_attr.json")
            self.assertEqual(df.loc[8, "slug"], "pasta-riso-cereali")
            self.assertEqual(df.loc[27, "slug"], "2-alt-pasta")
        except Exception as exc:
            raise exc

    def test_dataset_csv6_weird_index_with_attr(self):
        """ Test reading table with 'weird' index column explicitly marked in schema """
        try:
            df = self.read_dataframe_asset("ds_test_6_weird_index_with_attr.json")
            self.assertEqual(df.index.name, "indice")
            self.assertEqual(df.loc[8, "slug"], "pane-pasticceria")
            self.assertEqual(df.loc[27, "slug"], "sughi-scatolame-condimenti")
            self.assertEqual(df.loc[100598, "slug"], "2-alt-salumi")
        except Exception as exc:
            raise exc

    def test_dataset_csv7_timedelta(self):
        """ Test timespan to timedelta automatic conversion """
        try:
            df = self.read_dataframe_asset("ds_test_7_autoschema.json")

            # index is from column 'indice'
            self.assertEqual(df.loc[1, "elapsed"], pd.Timedelta("1 day"))
            self.assertEqual(df.loc[3, "elapsed"], pd.Timedelta("2 days"))
            self.assertEqual(df.loc[4, "elapsed"], pd.Timedelta("3 days"))
            self.assertEqual(df.loc[6, "elapsed"], pd.Timedelta("1 days 06:05:01.00003"))
        except Exception as exc:
            raise exc

    def test_dataset_csv7_autoschema(self):
        """ Test automatically generating an analitico schema from a pandas dataframe """
        try:
            df = self.read_dataframe_asset("ds_test_7_autoschema.json")
            schema = generate_schema(df)

            columns = schema["columns"]
            self.assertEqual(len(columns), 12)

            self.assertEqual(columns[0]["name"], "name")
            self.assertEqual(columns[0]["type"], "string")
            self.assertEqual(columns[1]["name"], "slug")
            self.assertEqual(columns[1]["type"], "category")
            self.assertEqual(columns[2]["name"], "parent_id")
            self.assertEqual(columns[2]["type"], "float")
            self.assertEqual(columns[3]["name"], "depth")
            self.assertEqual(columns[3]["type"], "integer")
            self.assertEqual(columns[4]["name"], "priority")
            self.assertEqual(columns[4]["type"], "integer")
            self.assertEqual(columns[5]["name"], "max_weight")
            self.assertEqual(columns[5]["type"], "integer")
            self.assertEqual(columns[6]["name"], "frozen")
            self.assertEqual(columns[6]["type"], "boolean")
            self.assertEqual(columns[7]["name"], "rate")
            self.assertEqual(columns[7]["type"], "float")
            self.assertEqual(columns[8]["name"], "has_ingredients_book")
            self.assertEqual(columns[8]["type"], "boolean")
            self.assertEqual(columns[9]["name"], "indice")
            self.assertEqual(columns[9]["type"], "integer")
            self.assertEqual(columns[9]["index"], True)
            self.assertEqual(columns[10]["name"], "updated_at")
            self.assertEqual(columns[10]["type"], "datetime")
            self.assertEqual(columns[11]["name"], "elapsed")
            self.assertEqual(columns[11]["type"], "timespan")
        except Exception as exc:
            raise exc

    def test_dataset_csv7_reordering(self):
        """ Test reordering of columns in the source """
        try:
            df = self.read_dataframe_asset("ds_test_7_reordering.json")
            self.assertEqual(len(df.columns), 12)
            self.assertEqual(df.columns[0], "depth")
            self.assertEqual(df.columns[1], "elapsed")
            self.assertEqual(df.columns[2], "frozen")
            self.assertEqual(df.columns[3], "has_ingredients_book")
            self.assertEqual(df.columns[4], "indice")
            self.assertEqual(df.columns[5], "max_weight")
            self.assertEqual(df.columns[6], "name")
            self.assertEqual(df.columns[7], "parent_id")
            self.assertEqual(df.columns[8], "priority")
            self.assertEqual(df.columns[9], "rate")
            self.assertEqual(df.columns[10], "slug")
            self.assertEqual(df.columns[11], "updated_at")
        except Exception as exc:
            raise exc

    def test_dataset_csv7_filtering(self):
        """ Test removing columns in the source """
        try:
            df = self.read_dataframe_asset("ds_test_7_filtering.json")
            self.assertEqual(len(df.columns), 4)
            self.assertEqual(df.columns[0], "indice")
            self.assertEqual(df.columns[1], "name")
            self.assertEqual(df.columns[2], "slug")
            self.assertEqual(df.columns[3], "frozen")
        except Exception as exc:
            raise exc

    def test_dataset_csv8_unicode(self):
        """ Test unicode chars in the source """
        try:
            df = self.read_dataframe_asset("ds_test_8_unicode.json")
            self.assertEqual(len(df.columns), 3)
            self.assertEqual(df.columns[0], "index")
            self.assertEqual(df.columns[1], "language")
            self.assertEqual(df.columns[2], "message")

            self.assertEqual(
                df.loc[0, "message"],
                "! \" # $ % & ' ( ) * + - . / 0 1 2 3 4 5 6 7 8 9 : ; < = > ? @ A B C D E F G H I J K L M N O P Q R S T U V W X Y Z [ \\ ] ^ _ ` a b c d e f g h i j k l m n o p q r s t u v w x y z { | } ~",
            )
            self.assertEqual(
                df.loc[1, "message"],
                "¡ ¢ £ ¤ ¥ ¦ § ¨ © ª « ¬ \xad ® ¯ ° ± ² ³ ´ µ ¶ · ¸ ¹ º » ¼ ½ ¾ ¿ À Á Â Ã Ä Å Æ Ç È É Ê Ë Ì Í Î Ï Ð Ñ Ò Ó Ô Õ Ö × Ø Ù Ú Û Ü Ý Þ ß à á â ã ä å æ ç è é ê ë ì í î ï ð ñ ò ó ô õ ö ÷ ø ù ú û ü ý þ ÿ",
            )
            self.assertEqual(
                df.loc[2, "message"],
                "Ё Ђ Ѓ Є Ѕ І Ї Ј Љ Њ Ћ Ќ Ў Џ А Б В Г Д Е Ж З И Й К Л М Н О П Р С Т У Ф Х Ц Ч Ш Щ Ъ Ы Ь Э Ю Я а б в г д е ж з и й к л м н о п р с т у ф х ц ч ш щ ъ ы ь э ю я ё ђ ѓ є ѕ і ї ј љ њ ћ ќ ў џ Ѡ ѡ Ѣ ѣ Ѥ ѥ Ѧ ѧ Ѩ ѩ Ѫ ѫ Ѭ ѭ Ѯ ѯ Ѱ ѱ Ѳ ѳ Ѵ ѵ Ѷ ѷ Ѹ ѹ Ѻ ѻ Ѽ ѽ Ѿ",
            )
            self.assertEqual(
                df.loc[3, "message"],
                "؛ ؟ ء آ أ ؤ إ ئ ا ب ة ت ث ج ح خ د ذ ر ز س ش ص ض ط ظ ع غ ـ ف ق ك ل م ن ه و ى ي ً ٌ ٍ َ ُ ِ ّ ْ ٠ ١ ٢ ٣ ٤ ٥ ٦ ٧ ٨ ٩ ٪ ٫ ٬ ٭ ٰ ٱ ٲ ٳ ٴ ٵ ٶ ٷ ٸ ٹ ٺ ٻ ټ ٽ پ ٿ ڀ ځ ڂ ڃ ڄ څ چ ڇ ڈ ډ ڊ ڋ ڌ ڍ ڎ ڏ ڐ ڑ ڒ ړ ڔ ڕ ږ ڗ ژ ڙ ښ ڛ ڜ ڝ ڞ ڟ ڠ ڡ ڢ ڣ ڤ ڥ ڦ ڧ ڨ ک",
            )
            self.assertEqual(
                df.loc[4, "message"],
                "ก ข ฃ ค ฅ ฆ ง จ ฉ ช ซ ฌ ญ ฎ ฏ ฐ ฑ ฒ ณ ด ต ถ ท ธ น บ ป ผ ฝ พ ฟ ภ ม ย ร ฤ ล ฦ ว ศ ษ ส ห ฬ อ ฮ ฯ ะ ั า ำ ิ ี ึ ื ุ ู ฺ ฿ เ แ โ ใ ไ ๅ ๆ ็ ่ ้ ๊ ๋ ์ ํ ๎ ๏ ๐ ๑ ๒ ๓ ๔ ๕ ๖ ๗ ๘ ๙ ๚ ๛",
            )
            self.assertEqual(
                df.loc[5, "message"],
                "一 丁 丂 七 丄 丅 丆 万 丈 三 上 下 丌 不 与 丏 丐 丑 丒 专 且 丕 世 丗 丘 丙 业 丛 东 丝 丞 丟 丠 両 丢 丣 两 严 並 丧 丨 丩 个 丫 丬 中 丮 丯 丰 丱 串 丳 临 丵 丶 丷 丸 丹 为 主 丼 丽 举 丿 乀 乁 乂 乃 乄 久 乆 乇 么 义 乊 之 乌 乍 乎 乏 乐 乑 乒 乓 乔 乕 乖 乗 乘 乙 乚 乛 乜 九 乞 也 习 乡 乢 乣 乤 乥 书 乧 乨 乩 乪 乫 乬 乭 乮 乯 买 乱 乲 乳 乴 乵 乶 乷 乸 乹 乺 乻 乼 乽 乾 乿",
            )
        except Exception as exc:
            raise exc

    def test_dataset_csv9_types_datetime_nulls(self):
        """ Test reading datetime that has null values """
        try:
            df = self.read_dataframe_asset("ds_test_9.json")
            self.assertEqual(df.dtypes[0], "int64")
            self.assertEqual(df.dtypes[1], "O")
            self.assertTrue(df.dtypes[2] == "datetime64[ns]")

            self.assertTrue(isinstance(df.iloc[0, 2], pd.Timestamp))
            self.assertTrue(isinstance(df.iloc[1, 2], pd.Timestamp))

            # all these are variantions on None and null
            self.assertTrue(pd.isnull(df.iloc[2, 2]))
            self.assertTrue(pd.isnull(df.iloc[3, 2]))
            self.assertTrue(pd.isnull(df.iloc[4, 2]))
            self.assertTrue(pd.isnull(df.iloc[5, 2]))
            self.assertTrue(pd.isnull(df.iloc[6, 2]))
        except Exception as exc:
            raise exc

    # TODO: test reading number that use . for thousands (eg: en-us, locale)

    # TODO: test datetime in localized formats

    # TODO: test missing values in various formats

    # TODO: test replacing missing values

    # TODO: test plugins pipeline

    # TODO: test modified values after plugin pipeline

    # TODO: test modified schema after plugin pipeline
